import numpy as np


def get_default_residuals_wrapper(model_function):
    """Return a wrapper around the default nanite residual function"""
    default_modeling_wrapper = get_default_modeling_wrapper(model_function)

    def default_residuals_wrapper(params, delta, force, weight_cp=5e-7):
        return residual(params=params,
                        delta=delta,
                        force=force,
                        model=default_modeling_wrapper,
                        weight_cp=weight_cp)

    return default_residuals_wrapper


def get_default_modeling_wrapper(model_function):
    """Return a wrapper around the default nanite modeling function"""
    def default_modeling_wrapper(params, delta):
        return model_direction_agnostic(model_function=model_function,
                                        params=params,
                                        delta=delta)
    return default_modeling_wrapper


def model_direction_agnostic(model_function, params, delta):
    """Call `model_function` while making sure that data are in correct order

    TODO: Re-evaluate usefulness of this method.
    """
    if delta[0] < delta[-1]:
        revert = True
    else:
        revert = False
    if revert:
        delta = delta[::-1]

    mf = model_function(delta=delta, **params.valuesdict())

    if revert:
        return mf[::-1]
    else:
        return mf


def residual(params, delta, force, model, weight_cp=5e-7):
    """Compute residuals for fitting

    Parameters
    ----------
    params: lmfit.Parameters
        The fitting parameters for `model`
    delta: 1D ndarray of lenght M
        The indentation distances
    force: 1D ndarray of length M
        The corresponding force data
    model: callable
        A model function accepting the arguments ``params``
        and ``delta``
    weight_cp: positive float or zero/False
        The distance from the contact point until which
        linear weights will be applied. Set to zero to
        disable weighting.
    """
    md = model(params, delta)
    resid = force - md

    if weight_cp:
        # weight the curve so that the data around the contact_point do
        # not affect the fit so much.
        weights = compute_contact_point_weights(
            cp=params["contact_point"].value,
            delta=delta,
            weight_dist=weight_cp)
        resid *= weights
    return resid


def compute_contact_point_weights(cp, delta, weight_dist=5e-7):
    """Compute contact point weights

    Parameters
    ----------
    cp: float
        Fitted contact point value
    delta: 1d ndarray of length N
        The indentation array along which weights will be computed.
    weight_width: float
        The distance from `cp` until which weights will be applied.

    Returns
    -------
    weights: 1d ndarray of length N
        The weights.

    Notes
    -----
    All variables should be given in the same units. The weights increase
    linearly from increasing distances of `delta-cp` from 0 to 1 and are
    1 outside of the weight width `abs(delta-cp)>weight_width`.
    """
    # weights are proportional to distance from contact point
    # normalized by weight_width.
    x = np.abs(delta-cp)
    x /= weight_dist
    x[x > 1] = 1
    return x



"""Test of models using expressions"""
import pathlib

import numpy as np

from nanite import IndentationGroup

from common import MockModelModuleExpr


data_dir = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_dir / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_expr_model():
    """Make sure that a model with an expression is fitted correctly"""
    # Reference fit
    rgrp = IndentationGroup(jpkfile)
    ridnt = rgrp[0]
    ridnt.apply_preprocessing(["compute_tip_position",
                               "correct_force_offset"])
    ridnt.fit_model(model_key="hertz_para",
                    params_initial=None,
                    x_axis="tip position",
                    y_axis="force",
                    weight_cp=False,
                    segment="retract")
    rparms = ridnt.fit_properties["params_fitted"]
    remod = rparms["E"].value

    with MockModelModuleExpr() as mod:
        grp = IndentationGroup(jpkfile)
        idnt = grp[0]
        idnt.apply_preprocessing(["compute_tip_position",
                                  "correct_force_offset"])
        idnt.fit_model(model_key=mod.model_key,
                       params_initial=None,
                       x_axis="tip position",
                       y_axis="force",
                       weight_cp=False,
                       segment="retract")
        parms = idnt.fit_properties["params_fitted"]
        # make sure the expression survives
        assert parms["E1"].expr == "virtual_parameter+E"
        emod = parms["E1"].value
        # There are some difference due to heuristics
        assert np.allclose(emod, remod, atol=0, rtol=3e-3)


def test_expr_model_limit():
    """Fit with a limit towards the correct solution"""
    # Reference fit
    rgrp = IndentationGroup(jpkfile)
    ridnt = rgrp[0]
    ridnt.apply_preprocessing(["compute_tip_position",
                               "correct_force_offset"])
    ridnt.fit_model(model_key="hertz_para",
                    params_initial=None,
                    x_axis="tip position",
                    y_axis="force",
                    weight_cp=False,
                    segment="retract")
    rparmsi = ridnt.fit_properties["params_initial"]

    with MockModelModuleExpr() as mod:
        grp = IndentationGroup(jpkfile)
        idnt = grp[0]
        idnt.apply_preprocessing(["compute_tip_position",
                                  "correct_force_offset"])
        params_initial = mod.get_parameter_defaults()
        params_initial["E"].set(value=19000)
        params_initial["virtual_parameter"].set(value=10, min=0, max=200)
        params_initial["contact_point"].set(
            value=rparmsi["contact_point"].value)
        idnt.fit_model(model_key=mod.model_key,
                       params_initial=params_initial,
                       x_axis="tip position",
                       y_axis="force",
                       weight_cp=False,
                       segment="retract")
        parms = idnt.fit_properties["params_fitted"]
        # make sure the expression survives
        assert parms["E1"].expr == "virtual_parameter+E"
        emod = parms["E1"].value
        # It should have gone to the boundary
        assert np.allclose(emod, 19200)


def test_expr_model_sign():
    """Fit with negative limit"""
    # Reference fit
    rgrp = IndentationGroup(jpkfile)
    ridnt = rgrp[0]
    ridnt.apply_preprocessing(["compute_tip_position",
                               "correct_force_offset"])
    ridnt.fit_model(model_key="hertz_para",
                    params_initial=None,
                    x_axis="tip position",
                    y_axis="force",
                    weight_cp=False,
                    segment="retract")
    rparms = ridnt.fit_properties["params_fitted"]
    rparmsi = ridnt.fit_properties["params_initial"]
    remod = rparms["E"].value

    with MockModelModuleExpr() as mod:
        grp = IndentationGroup(jpkfile)
        idnt = grp[0]
        idnt.apply_preprocessing(["compute_tip_position",
                                  "correct_force_offset"])
        params_initial = mod.get_parameter_defaults()
        params_initial["E"].set(value=20000)
        params_initial["virtual_parameter"].set(value=-1, min=-np.inf, max=0)
        params_initial["contact_point"].set(
            value=rparmsi["contact_point"].value)
        idnt.fit_model(model_key=mod.model_key,
                       params_initial=params_initial,
                       x_axis="tip position",
                       y_axis="force",
                       weight_cp=False,
                       segment="retract")
        parms = idnt.fit_properties["params_fitted"]
        # make sure the expression survives
        assert parms["E1"].expr == "virtual_parameter+E"
        emod = parms["E1"].value
        # There are some difference due to heuristics
        assert np.allclose(emod, remod, atol=0, rtol=3e-3)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import argparse
import io
from functools import lru_cache
import getpass
from os import fspath
import pathlib
try:
    import tkinter as tk
except (ImportError, FileNotFoundError):
    TK_AVAILABLE = False
else:
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
    TK_AVAILABLE = True

import types

import afmformats
from matplotlib.figure import Figure
import matplotlib.image as mpimg
import tifffile

from ..indent import Indentation
from ..group import IndentationGroup
from ..read import get_data_paths_enum
from ..rate import io as rio

from .profile import Profile, PROFILE_PATH
from .plotting import plot_data


class RatingGUI:
    def __init__(self, root, data_paths, h5path):
        self.data_paths = data_paths
        self.h5path = h5path
        self.current_index = 0
        self.idnt = None
        self.root = root

        root.wm_protocol("WM_DELETE_WINDOW", self.close)
        tk.Grid.rowconfigure(root, 0, weight=1)
        tk.Grid.columnconfigure(root, 0, weight=1)

        root.title("nanite force-distance rater")

        master = tk.Frame(root)
        master.grid(row=0, column=0, sticky=tk.N+tk.S+tk.E+tk.W)
        self.master = master

        tk.Grid.rowconfigure(master, 1, weight=1)

        for ii in range(7):
            tk.Grid.columnconfigure(master, ii, weight=1)

        l1 = tk.Label(master, text="User text!")
        l1.grid(row=0, column=0, columnspan=6)

        self.figure = Figure(figsize=(10, 7), dpi=100)
        self.canvas = FigureCanvasTkAgg(self.figure, master=master)
        self.canvas.get_tk_widget().grid(row=1,
                                         column=0,
                                         columnspan=6,
                                         sticky=tk.N+tk.S+tk.E+tk.W)

        l2 = tk.Label(master, text="user:")
        l2.grid(row=2, column=0, sticky=tk.E)
        self.user = tk.Entry(master, text="user:", width=30)
        self.user.insert(0, getpass.getuser())
        self.user.grid(row=2, column=1, sticky=tk.W)

        l3 = tk.Label(master, text="rating:")
        l3.grid(row=2, column=2, sticky=tk.E)
        self.rating = tk.Spinbox(master, from_=-1, to=10)
        self.rating.grid(row=2, column=3, sticky=tk.W)

        self.prev_button = tk.Button(master,
                                     text="Previous (Alt+Left)",
                                     command=self.previous)
        root.bind('<Alt-Left>', self.previous)
        self.prev_button.grid(row=2, column=4, sticky=tk.W+tk.E)

        self.next_button = tk.Button(master,
                                     text="Next (Alt+Right)",
                                     command=self.next)
        root.bind('<Alt-Right>', self.next)
        self.next_button.grid(row=2, column=5, sticky=tk.W+tk.E)

        l4 = tk.Label(master, text="comment:")
        l4.grid(row=3, column=0, sticky=tk.E)
        self.comment = tk.Entry(master, width=100)
        self.comment.grid(row=3, column=1, columnspan=5, sticky=tk.W+tk.E)

        # skip to first unrated curve
        while True:
            fakeidnt = types.SimpleNamespace()
            fakeidnt.path = self.current_path
            fakeidnt.enum = self.current_enum
            rated, _, _ = rio.hdf5_rated(self.h5path, indent=fakeidnt)
            if not rated:
                break
            else:
                self.current_index += 1
        self.load()

    @property
    def current_path(self):
        return self.data_paths[self.current_index][0]

    @property
    def current_enum(self):
        return self.data_paths[self.current_index][1]

    def close(self):
        self.save()
        self.root.destroy()

    def next(self, e=None):
        self.save()
        self.current_index += 1
        if self.current_index == len(self.data_paths):
            self.current_index = 0
        self.load()

    def previous(self, e=None):
        self.save()
        self.current_index -= 1
        if self.current_index < 0:
            self.current_index = len(self.data_paths) - 1
        self.load()

    def load(self):
        self.set_label_path()
        path = self.current_path
        enum = self.current_enum
        self.idnt = fit_data(path=path, enum=enum)
        plot_data(self.idnt, figure=self.figure)
        self.canvas.draw()

        # set rating
        self.rating.delete(0, tk.END)
        self.comment.delete(0, tk.END)
        rated, rating, comment = rio.hdf5_rated(self.h5path, indent=self.idnt)
        if rated:
            self.rating.insert(0, "{}".format(rating))
            self.comment.insert(0, "{}".format(comment))

    def save(self):
        rating = self.rating.get()
        if rating:
            rio.save_hdf5(h5path=self.h5path,
                          indent=self.idnt,
                          user_rate=int(rating),
                          user_name=self.user.get(),
                          user_comment=self.comment.get(),
                          h5mode="a")

    def set_label_path(self):
        if hasattr(self, "label_path"):
            self.label_path.destroy()
        self.label_path = tk.Label(self.master,
                                   text="curve {}/{}: {} - {}".format(
                                       self.current_index + 1,
                                       len(self.data_paths),
                                       self.current_path,
                                       self.current_enum))
        self.label_path.grid(row=0, column=0, columnspan=6)


def fit():
    parser = fit_parser()
    args = parser.parse_args()
    path = pathlib.Path(args.data_path).resolve()
    pout = pathlib.Path(args.out_dir).resolve()
    pout.mkdir(exist_ok=True, parents=True)
    fit_perform(path, path_results=pout)


def fit_perform(path, path_results, profile_path=PROFILE_PATH):
    path_results = pathlib.Path(path_results)
    ptsv = path_results / "statistics.tsv"
    ptif = path_results / "plots.tif"
    # exported data columns
    pf = Profile(path=profile_path, create=False)
    dlist = [["path", lambda x: x.path],
             ["enum", lambda x: x.enum],
             ["E", lambda x: x.fit_properties["params_fitted"]["E"].value],
             ["rating", lambda x: round(x.rate_quality(
                 training_set=pf["rating training set"],
                 regressor=pf["rating regressor"]),
                 ndigits=1)],
             ]
    ddict = dict(dlist)
    header = "\t".join([dd[0] for dd in dlist])
    with ptsv.open(mode="w") as ts:
        ts.write(header + "\n")
    # get all files in path
    datapaths = afmformats.find_data(path, modality="force-distance")
    with tifffile.TiffWriter(fspath(ptif), imagej=True) as tf, \
            ptsv.open(mode="a") as ts:
        for pp in datapaths:
            print("Processing: {}".format(pp))
            grp = IndentationGroup(pp)
            for idnt in grp:
                fit_data(idnt, profile_path=profile_path)
                # save statistics
                stats = [str(dd[1](idnt)) for dd in dlist]
                ts.write("\t".join(stats) + "\n")
                # save plot
                imio = io.BytesIO()
                rating_text = "Rating parameters:\n" \
                    + "regressor: {}\n".format(pf["rating regressor"]) \
                    + "training set: {}\n".format(pf["rating training set"]) \
                    + "rating: {:.1f}\n".format(ddict["rating"](idnt))
                plot_data(idnt,
                          add_text=rating_text,
                          path=imio)
                imio.seek(0)
                imdat = (mpimg.imread(imio) * 255).astype("uint8")
                tf.write(imdat, contiguous=True)


@lru_cache(maxsize=5)
def fit_data(path, enum=0, profile_path=PROFILE_PATH):
    if isinstance(path, Indentation):
        idnt = path
    else:
        idnt = IndentationGroup(path)[enum]

    pf = Profile(path=profile_path, create=False)

    idnt.apply_preprocessing(preprocessing=pf["preprocessing"],
                             options=pf["preprocessing_options"])
    params = pf.get_fit_params()

    idnt.fit_model(model_key=pf["model_key"],
                   params_initial=params,
                   range_type=pf["range_type"],
                   range_x=pf["range_x"],
                   segment=pf["segment"],
                   weight_cp=pf["weight_cp"],
                   )
    return idnt


def fit_parser():
    descr = "Fit AFM force-distance data. Statistics (.tsv file) and " \
            + "visualizations of the fits (multi-page .tif file) are stored " \
            + "in the results directory."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('data_path', type=str,
                        help='input folder containing AFM force-distance '
                             + 'data')
    parser.add_argument('out_dir', type=str,
                        help='results directory')
    return parser


def generate_training_set_parser():
    descr = "Create a training set for usage in nanite from rating " \
            "containers (.h5 files manually created with `nanite-rate`)."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('data_path', type=str,
                        help='path to a rating container or a folder '
                             + 'containing rating containers')
    parser.add_argument('out_dir', type=str,
                        help='directory where the training set will be '
                             ' stored')
    return parser


def generate_training_set():
    parser = generate_training_set_parser()
    args = parser.parse_args()
    data_path = pathlib.Path(args.data_path).resolve()
    out_dir = pathlib.Path(args.out_dir).resolve()
    name = data_path.stem
    if data_path.exists():
        print("Generating training set '{}'.".format(name))
        rm = rio.RateManager(data_path, verbose=1)
        rm.export_training_set(out_dir / "ts_{}".format(name))
    else:
        print("Path does not exist: '{}'".format(data_path))


def rate():
    parser = rate_parser()
    args = parser.parse_args()
    path = pathlib.Path(args.data_path).resolve()
    h5 = pathlib.Path(args.rating_path).resolve()
    if not h5.name.endswith(".h5"):
        h5 = h5.with_name(h5.name + ".h5")
    # get all files in path
    enumpaths = get_data_paths_enum(path, skip_errors=True)
    # for now, only compare single curves
    enumpaths = [ee for ee in enumpaths if ee[0].suffix == ".jpk-force"]
    if TK_AVAILABLE:
        # start GUI
        root = tk.Tk()
        RatingGUI(root, data_paths=enumpaths, h5path=h5)
        root.mainloop()
    else:
        raise ValueError("tkinter is not available! (frozen application?)")


def rate_parser():
    descr = "Manually rate (the fit to) AFM force-distance data. " \
            + "A graphical user interface allows to rate and comment on " \
            + "each force-distance curve. The fits and the raw data are " \
            + "stored in a rating container that can then be passed to " \
            + "`nanite-generate-training-set`."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('data_path', type=str,
                        help='input folder containing AFM force-distance '
                             + 'data')
    parser.add_argument('rating_path', type=str,
                        help='path to the output rating container (will '
                             + 'be created if it does not already exist)')
    return parser



from collections import OrderedDict
import inspect
import warnings

import numpy as np

from . import residuals


class ModelError(BaseException):
    pass


class ModelIncompleteError(ModelError):
    pass


class ModelImplementationError(ModelError):
    pass


class ModelImportError(ModelError):
    pass


class ModelImplementationWarning(UserWarning):
    pass


class NaniteFitModel:
    def __init__(self, model_module):
        """Initialize the model with an imported Python module"""
        self.module = model_module
        self._module_check()
        self._module_autocomplete()

        # propagate all module attributes to this instance
        # standard parameters
        self.get_parameter_defaults = self.module.get_parameter_defaults
        self.model_doc = self.module.model_doc
        self.model_key = self.module.model_key
        self.model_name = self.module.model_name
        self.parameter_keys = self.module.parameter_keys
        self.parameter_names = self.module.parameter_names
        self.parameter_units = self.module.parameter_units
        self.valid_axes_x = self.module.valid_axes_x
        self.valid_axes_y = self.module.valid_axes_y
        # optional ancillary parameters
        if hasattr(self.module, "compute_ancillaries"):
            self.has_module_ancillaries = True
            self.parameter_anc_keys = self.module.parameter_anc_keys
            self.parameter_anc_names = self.module.parameter_anc_names
            self.parameter_anc_units = self.module.parameter_anc_units
        else:
            self.has_module_ancillaries = False
        # model function
        self.model = self.module.model
        # residuals
        self.residual = self.module.residual

    def __str__(self):
        return f"NaniteFitModel '{self.model_key}'"

    def __repr__(self):
        return f"<NaniteFitModel '{self.model_key}' at {hex(id(self))}"

    def _module_autocomplete(self):
        """Add any missing attributes to the underlying model module"""
        # check for residuals function
        if not hasattr(self.module, "residual"):
            # use the default residual function
            self.module.residual = residuals.get_default_residuals_wrapper(
                model_function=self.module.model_func)

        # check for modeling function
        if not hasattr(self.module, "model"):
            # use the default residual function
            self.module.model = residuals.get_default_modeling_wrapper(
                model_function=self.module.model_func)

    def _module_check(self):
        """Checks whether the model's module is set up correctly"""
        # sanity checks
        missing = []
        for attr in [
            "get_parameter_defaults",
            "model_doc",
            "model_key",
            "model_name",
            "parameter_keys",
            "parameter_names",
            "parameter_units",
            "valid_axes_x",
            "valid_axes_y",
             ]:
            if not hasattr(self.module, attr):
                missing.append(attr)
        if missing:
            raise ModelIncompleteError(
                f"Model `{self.module}` is missing the following "
                + f"attributes: {', '.join(missing)}")

        model_key = self.module.model_key
        # check for completeness of ancillary parameter recipe
        if hasattr(self.module, "compute_ancillaries"):
            missing_anc = []
            for attr in ["parameter_anc_keys",
                         "parameter_anc_names",
                         "parameter_anc_units",
                         ]:
                if not hasattr(self.module, attr):
                    missing_anc.append(attr)
            if missing_anc:
                raise ModelIncompleteError(
                    f"Model `{model_key}` is missing the following "
                    + f"attributes: {', '.join(missing_anc)}")

        # check length of modeling lists
        if len(self.module.parameter_keys) != len(self.module.parameter_names):
            raise ModelImplementationError(
                "'parameter_keys' and 'parameter_names' have different "
                + f"lengths for model '{model_key}'!")
        if len(self.module.parameter_keys) != len(self.module.parameter_units):
            raise ModelImplementationError(
                "'parameter_keys' and 'parameter_units' have different "
                + f"lengths for model '{model_key}'!")

        # check for spaces in units
        if [u.strip() for u in self.module.parameter_units] \
                != self.module.parameter_units:
            warnings.warn("The `parameter_units` should not contain leading "
                          + f"or trailing spaces. Please check {model_key}!",
                          ModelImplementationWarning)

        if hasattr(self.module, "parameter_anc_units"):
            if [u.strip() for u in self.module.parameter_anc_units] \
                    != self.module.parameter_anc_units:
                warnings.warn(
                    "The `parameter_anc_units` should not contain leading "
                    + f"or trailing spaces. Please check {model_key}!",
                    ModelImplementationWarning)

        # check for label uniqueness
        if len(self.module.parameter_names) \
                != len(set(self.module.parameter_names)):
            raise ModelImplementationError(
                f"'parameter_names' should be unique for '{model_key}'!")

        # checks for model parameters
        p_def = list(self.module.get_parameter_defaults().keys())
        p_arg = list(inspect.signature(
            self.module.model_func).parameters.keys())
        for ii, key in enumerate(self.module.parameter_keys):
            if key != p_def[ii]:
                raise ModelImplementationError(
                    "Please check 'parameter_keys' and "
                    + f"'get_parameter_defaults'  of the model '{model_key}'. "
                    + f"Keys {key} and {p_def[ii]} are not in order!")
            if key != p_arg[ii+1]:
                warnings.warn(
                    "Please make sure that the parameters of the model "
                    + "function are in the same order as in 'parameter_keys' "
                    + f"for the model '{model_key}'! "
                    + "The abscissa (usually `delta`) should come first. "
                    + "This warning may become an Exception in the future!",
                    ModelImplementationWarning)

    def compute_ancillaries(self, fd):
        """Compute ancillary parameters for a force-distance dataset

        Ancillary parameters include parameters that:

        - are unrelated to fitting: They may just be important parameters
          to the user.
        - require the entire dataset: They cannot be extracted during
          fitting, because they require more than just the approach xor
          retract curve to compute (e.g. hysteresis, jump of retract curve
          at maximum indentation). They may, additionally, depend on
          initial fit parameters set by the user.
        - require a fit: They are dependent on fitting parameters but
          are not required during fitting.

        Notes
        -----
        If an ancillary parameter name matches that of a fitting parameter,
        then it is assumed that it can be used for fitting. Please see
        :func:`nanite.indent.Indentation.get_initial_fit_parameters`
        and :func:`nanite.fit.guess_initial_parameters`.

        Ancillary parameters are set to `np.nan` if they cannot be
        computed.

        Parameters
        ----------
        fd: nanite.indent.Indentation
            The force-distance data for which to compute the ancillary
            parameters

        Returns
        -------
        ancillaries: collections.OrderedDict
            key-value dictionary of ancillary parameters
        """
        # TODO:
        # - ancillaries are not cached yet (some ancillaries might depend on
        #   fitting interval or other initial parameters - take that into
        #   account)
        # - "max_indent" actually belongs to "common_ancillaries" (see fit.py)
        anc_ord = OrderedDict()
        # general
        for key in ANCILLARY_COMMON:
            gmeth = ANCILLARY_COMMON[key][2]
            anc_ord[key] = gmeth(fd)
        # from module
        if self.has_module_ancillaries:
            anc_md = self.module.compute_ancillaries(fd)
            for kk in self.parameter_anc_keys:
                anc_ord[kk] = anc_md[kk]
        return anc_ord

    def get_anc_parm_keys(self):
        """Return the key names of a model's ancillary parameters"""
        akeys = list(ANCILLARY_COMMON.keys())
        if self.has_module_ancillaries:
            akeys += self.parameter_anc_keys
        return akeys

    def get_parm_name(self, key):
        """Return parameter label

        Parameters
        ----------
        key: str
            The parameter key (e.g. "E")

        Returns
        -------
        parm_name: str
            The parameter label (e.g. "Young's Modulus")
        """
        if key in self.parameter_keys:
            idx = self.parameter_keys.index(key)
            return self.parameter_names[idx]
        elif (self.has_module_ancillaries
              and key in self.parameter_anc_keys):
            idx = self.parameter_anc_keys.index(key)
            return self.parameter_anc_names[idx]
        elif key in ANCILLARY_COMMON:
            return ANCILLARY_COMMON[key][0]
        else:
            raise KeyError(
                f"Could not find parameter name for '{key}' in '{self}'!")

    def get_parm_unit(self, key):
        """Return parameter unit

        Parameters
        ----------
        key: str
            The parameter key (e.g. "E")

        Returns
        -------
        parm_unit: str
            The parameter unit (e.g. "Pa")
        """
        if key in self.parameter_keys:
            idx = self.parameter_keys.index(key)
            return self.parameter_units[idx]
        elif (self.has_module_ancillaries
              and key in self.parameter_anc_keys):
            idx = self.parameter_anc_keys.index(key)
            return self.parameter_anc_units[idx]
        elif key in ANCILLARY_COMMON:
            return ANCILLARY_COMMON[key][1]
        else:
            raise KeyError(
                f"Could not find parameter unit for '{key}' in '{self}'!")


def compute_anc_max_indent(fd):
    """Compute ancillary parameter 'Maximum indentation'"""
    # compute maximal indentation
    if ("tip position" in fd
        and "fit" in fd
            and "params_fitted" in fd.fit_properties
            and "contact_point" in fd.fit_properties["params_fitted"]):
        cp = fd.fit_properties["params_fitted"]["contact_point"].value
        idmax = fd.appr["fit"].argmax()
        mi = fd.appr["tip position"][idmax]
        mival = (cp-mi)
    else:
        mival = np.nan
    return mival


#: Common ancillary parameters
ANCILLARY_COMMON = OrderedDict()
ANCILLARY_COMMON["max_indent"] = ("Maximum indentation",
                                  "m",
                                  compute_anc_max_indent)



"""Test of ancillary parameters"""
import pathlib

import numpy as np

import nanite
import nanite.model

from common import MockModelModule


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_simple_ancillary_override():
    """basic test for ancillary parameters"""
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]

    with MockModelModule(
        compute_ancillaries=lambda x: {"E": 1580},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):
        # We need to perform preprocessing first, if we want to get the
        # correct initial contact point.
        idnt.apply_preprocessing(["compute_tip_position"])
        # We set the baseline fixed, because this test was written so)
        params_initial = idnt.get_initial_fit_parameters(model_key="test1")
        params_initial["baseline"].set(vary=False)
        idnt.fit_model(model_key="test1",
                       params_initial=params_initial)
        assert idnt.fit_properties["params_initial"]["E"].value == 1580
        assert np.allclose(idnt.fit_properties["params_fitted"]["E"].value,
                           1584.8941261696934,
                           atol=1,
                           rtol=0)


def test_simple_ancillary_override_nan():
    """nan values are not used and should be ignored"""
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]

    with MockModelModule(
        compute_ancillaries=lambda x: {"E": np.nan},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test2"):
        # We need to perform preprocessing first, if we want to get the
        # correct initial contact point.
        idnt.apply_preprocessing(["compute_tip_position"])
        # We set the baseline fixed, because this test was written so)
        params_initial = idnt.get_initial_fit_parameters(model_key="test2")
        params_initial["baseline"].set(vary=False)
        idnt.fit_model(model_key="test2",
                       params_initial=params_initial)
        assert idnt.fit_properties["params_initial"]["E"].value == 3000
        assert np.allclose(idnt.fit_properties["params_fitted"]["E"].value,
                           1584.8876592662375,
                           atol=0,
                           rtol=1e-5)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import copy
import inspect
import functools
import warnings

import numpy as np

from . import poc
from .smooth import smooth_axis_monotone


#: Available preprocessors
PREPROCESSORS = []


class CannotSplitWarning(UserWarning):
    pass


def apply(apret, identifiers=None, options=None, ret_details=False,
          preproc_names=None):
    """Perform force-distance preprocessing steps

    Parameters
    ----------
    apret: nanite.Indentation
        The afm data to preprocess
    identifiers: list
        A list of preprocessing identifiers that will be
        applied (in the order given).
    options: dict of dict
        Preprocessing options for each identifier
    ret_details:
        Return preprocessing details dictionary
    preproc_names: list
        Deprecated - use `identifiers` instead
    """
    if preproc_names is not None:
        identifiers = preproc_names
        warnings.warn(
            "Please use 'identifiers' instead of 'preproc_names'!",
            DeprecationWarning)
    details = {}
    # Reset all user-defined data of the dataset, because we
    # probably edited "tip position", "force", etc. If this
    # is too slow in the future, then we would have to remember
    # the previous preprocessing pipeline (such code is currently
    # implemented on the other side in `indent.py` (2021-08-16).
    apret.reset_data()
    for ii, pid in enumerate(identifiers):
        if pid in available():
            meth = get_func(pid)
            req = meth.steps_required
            act = identifiers[:ii]
            if req is not None and ((set(req) & set(act)) != set(req)):
                raise ValueError(f"The preprocessing step '{pid}' requires"
                                 f" the steps {meth.steps_required}!")
            # create a copy of the dictionary (if it exists) so that
            # `ret_details` is not written to it
            kwargs = copy.deepcopy(options.get(pid, {}))
            if "ret_details" in inspect.signature(meth).parameters:
                # only set `ret_details` if method accepts it
                kwargs["ret_details"] = ret_details
            details[pid] = meth(apret, **kwargs)
        else:
            msg = "The preprocessing method '{}' does not exist!"
            raise KeyError(msg.format(pid))
    # only return details if required
    return details if ret_details else None


def autosort(identifiers):
    """Automatically sort preprocessing identifiers

    This takes into account `steps_required` and `steps_optional`.
    """
    sorted_identifiers = copy.copy(identifiers)
    for pid in identifiers:
        meth = get_func(pid)
        steps_precursor = []
        if meth.steps_required is not None:
            steps_precursor += meth.steps_required
        if meth.steps_optional is not None:
            for ostep in meth.steps_optional:
                if ostep in identifiers:
                    steps_precursor.append(ostep)
        for step in steps_precursor:
            # We have a requirement, check whether it is fulfilled
            cix = sorted_identifiers.index(pid)
            rix = sorted_identifiers.index(step)
            if rix > cix:
                # We pop the wrong requirement and insert it before
                # the current pid.
                sorted_identifiers.remove(step)
                sorted_identifiers.insert(cix, step)

    # Perform a sanity check
    check_order(sorted_identifiers)

    return sorted_identifiers


@functools.lru_cache()
def available():
    """Return list of available preprocessor identifiers"""
    av = [pp.identifier for pp in PREPROCESSORS]
    return autosort(av)


def check_order(identifiers):
    """Check preprocessing steps for correct order"""
    for cix, pid in enumerate(identifiers):
        meth = get_func(pid)
        if meth.steps_required:
            rix = [identifiers.index(r) for r in meth.steps_required]
            if np.any(np.array(rix) > cix):
                raise ValueError(
                    f"Wrong required step order for {pid}: {identifiers}!")
        if meth.steps_optional:
            rio = []
            for rr in meth.steps_optional:
                if rr in identifiers:
                    rio.append(identifiers.index(rr))
            if np.any(np.array(rio) > cix):
                raise ValueError(
                    f"Wrong optional step order for {pid}: {identifiers}!")


def get_func(identifier):
    """Return preprocessor function for identifier"""
    for func in PREPROCESSORS:
        if func.identifier == identifier:
            return func
    else:
        raise KeyError(f"Preprocessor '{identifier}' unknown!")


def get_name(identifier):
    """Return preprocessor name for identifier"""
    func = get_func(identifier)
    return func.name


def get_steps_required(identifier):
    """Return requirement identifiers for identifier"""
    func = get_func(identifier)
    return func.steps_required


def preprocessing_step(identifier, name, steps_required=None,
                       steps_optional=None, options=None):
    """Decorator for Indentation preprocessors

    The name and identifier are stored as a property of the wrapped
    function.

    Parameters
    ----------
    identifier: str
        identifier of the preprocessor (e.g. "correct_tip_offset")
    name: str
        human-readble name of the preprocessor
        (e.g. "Estimate contact point")
    steps_required: list of str
        list of preprocessing steps that must be added before this
        step
    steps_optional: list of str
        unlike `steps_required`, these steps do not have to be set,
        but if they are set, they should come before this step
    options: list of dict
        if the preprocessor accepts optional keyword arguments,
        this list yields valid values or dtypes
    """

    def attribute_setter(func):
        """Decorator that sets the necessary attributes

        The outer decorator is used to obtain the attributes.
        This inner decorator returns the actual function that
        wraps the preprocessor.
        """
        func.identifier = identifier
        assert isinstance(name, str)
        func.name = name
        func.options = options
        func.steps_required = steps_required
        func.steps_optional = steps_optional
        PREPROCESSORS.append(func)
        return func

    return attribute_setter


@preprocessing_step(identifier="compute_tip_position",
                    name="tip-sample separation")
def preproc_compute_tip_position(apret):
    """Perform tip-sample separation

    Populate the "tip position" column by adding the force
    normalized by the spring constant to the cantilever
    height ("height (measured)").

    This computation correctly reproduces the column
    "Vertical Tip Position" as it is exported by the
    JPK analysis software with the checked option
    "Use Unsmoothed Height".
    """
    has_hm = "height (measured)" in apret
    has_fo = "force" in apret
    has_sc = "spring constant" in apret.metadata
    if "tip position" in apret.columns_innate:
        # nothing to do
        pass
    elif has_hm and has_fo and has_sc:
        k = apret.metadata["spring constant"]
        force = apret["force"]
        zcant = apret["height (measured)"]
        apret["tip position"] = zcant + force / k
    else:
        missing = []
        if not has_hm:
            missing.append("missing data column 'height (measured)'")
        if not has_fo:
            missing.append("missing data column 'force'")
        if not has_sc:
            missing.append("missing metadata 'spring constant'")
        mt = ", ".join(missing)
        raise ValueError("Cannot compute tip position: {}".format(mt))


@preprocessing_step(identifier="correct_force_offset",
                    name="baseline correction")
def preproc_correct_force_offset(apret):
    """Correct the force offset with an average baseline value
    """
    idp = poc.compute_poc(force=apret["force"],
                          method="deviation_from_baseline")
    if idp:
        apret["force"] = apret["force"] - np.average(apret["force"][:idp])
    else:
        apret["force"] = apret["force"] - apret["force"][0]


@preprocessing_step(
    identifier="correct_tip_offset",
    name="contact point estimation",
    steps_required=["compute_tip_position"],
    options=[
        {"name": "method",
         "type": str,
         "choices": [p.identifier for p in poc.POC_METHODS],
         "choices_human_readable": [p.name for p in poc.POC_METHODS]}
    ]
)
def preproc_correct_tip_offset(apret, method="deviation_from_baseline",
                               ret_details=False):
    """Estimate the point of contact

    An estimate of the contact point is subtracted from the
    tip position.
    """
    data = poc.compute_poc(force=apret["force"],
                           method=method,
                           ret_details=ret_details)
    if ret_details:
        cpid, details = data
    else:
        cpid, details = data, None
    apret["tip position"] = (apret["tip position"]
                             - apret["tip position"][cpid])
    return details


@preprocessing_step(identifier="correct_split_approach_retract",
                    name="segment discovery",
                    steps_required=["compute_tip_position"])
def preproc_correct_split_approach_retract(apret):
    """Split the approach and retract curves (farthest point method)

    Approach and retract curves are defined by the microscope. When the
    direction of piezo movement is flipped, the force at the sample tip
    is still increasing. This can be either due to a time lag in the AFM
    system or due to a residual force acting on the sample due to the
    bent cantilever.

    To repair this time lag, we append parts of the retract curve to the
    approach curve, such that the curves are split at the minimum height.
    """
    x = np.array(apret["tip position"], copy=True)
    y = np.array(apret["force"], copy=True)

    idp = poc.poc_deviation_from_baseline(y)
    if idp and not np.isnan(idp):
        # Flip and normalize tip position so that maximum is at minimum
        # z-position (set to 1) which coincides with maximum indentation.
        x -= x[idp]
        x /= x.min()
        x[x < 0] = 0

        # Flip and normalize force so that maximum force is set to 1.
        y -= np.average(y[:idp])
        y /= y.max()
        y[y < np.std(y[:idp])] = 0

        idmin = np.argmax(x ** 2 + y ** 2)

        segment = np.zeros(len(apret), dtype=np.uint8)
        segment[idmin:] = 1
        apret["segment"] = segment
    else:
        msg = "Cannot correct splitting of approach and retract curve " + \
              "because the contact point position could not be estimated."
        warnings.warn(msg, CannotSplitWarning)


@preprocessing_step(identifier="smooth_height",
                    name="monotonic height data",
                    steps_optional=[
                        # Otherwise we lose the location of the point
                        # of deepest indentation:
                        "correct_split_approach_retract",
                        # Otherwise it might not be applied to
                        # "tip position":
                        "compute_tip_position"])
def preproc_smooth_height(apret):
    """Make height data monotonic

    For the columns "height (measured)", "height (piezo), and
    "tip position", this method ensures that the approach and
    retract segments are monotonic.
    """
    orig = ["height (measured)",
            "height (piezo)",
            "tip position"]
    for col in orig:
        if col not in apret:
            continue
        # Apply smoothing
        sm_app = smooth_axis_monotone(apret.appr[col])
        sm_ret = smooth_axis_monotone(apret.retr[col])

        # Replace the column data
        apret.appr[col] = sm_app
        apret.retr[col] = sm_ret


def _deprecate_call(method):
    def wrapper(*args, **kwargs):
        warnings.warn(
            "Using `IndentationPreprocessor` class is deprecated, please use "
            f"the method '{method.__module__}.{method.__name__}' instead!",
            DeprecationWarning)
        return method(*args, **kwargs)
    return wrapper


class IndentationPreprocessor:
    apply = _deprecate_call(apply)
    autosort = _deprecate_call(autosort)
    get_func = _deprecate_call(get_func)
    get_name = _deprecate_call(get_name)
    get_steps_required = _deprecate_call(get_steps_required)
    check_order = _deprecate_call(check_order)
    available = _deprecate_call(available)



# cython: language_level=3
# cython: binding=True
import lmfit
import numpy as np
cimport numpy as np


def get_parameter_defaults():
    """Return the default model parameters"""
    # The order of the parameters must match the order 
    # of ´parameter_names´ and ´parameter_keys´.
    params = lmfit.Parameters()
    params.add("E", value=3e3, min=0)
    params.add("R", value=10e-6, min=0, vary=False)
    params.add("nu", value=.5, min=0, max=0.5, vary=False)
    params.add("contact_point", value=0)
    params.add("baseline", value=0)
    return params


def hertz_spherical(delta, double E, double R, double nu,
                    double contact_point=0, double baseline=0):
    r"""Hertz model for Spherical indenter - modified by Sneddon


    .. math::

        F &= \frac{E}{1-\nu^2} \left( \frac{R^2+a^2}{2} \ln \! \left(
             \frac{R+a}{R-a}\right) -aR  \right)\\
        \delta &= \frac{a}{2} \ln
             \! \left(\frac{R+a}{R-a}\right)

    (:math:`a` is the radius of the circular contact area between bead
    and sample.)

    Parameters
    ----------
    delta: 1d ndarray
        Indentation [m]
    E: float
        Young's modulus [N/m²]
    R: float
        Tip radius [m]
    nu: float
        Poisson's ratio
    contact_point: float
        Indentation offset [m]
    baseline: float
        Force offset [N]

    Returns
    -------
    F: float
        Force [N]

    Notes
    -----
    These approximations are made by the Hertz model:

    - The sample is isotropic.
    - The sample is a linear elastic solid.
    - The sample is extended infinitely in one half space.
    - The indenter is not deformable.
    - There are no additional interactions between sample and indenter.

    Additional assumptions:

    - no surface forces

    References
    ----------
    Sneddon (1965) :cite:`Sneddon1965` (equations 6.13 and 6.15)
    """
    cdef double a
    cdef int ii
    FF = np.zeros_like(delta)
    root = delta-contact_point
    root[root >0]=0
    root=-1*root
    for ii in range(root.shape[0]):
        if root[ii]==0:
            FF[ii]=0
        else:
            a=_get_a(R, root[ii])
            FF[ii]=E/(1-nu**2)*((R**2+a**2)/2*np.log((R+a)/(R-a))-a*R) 
    return FF + baseline


cdef double _get_a(double R, double delta, double accuracy=1e-12):
    cdef double a_left, a_center, a_right, d_delta, delta_predict
    a_left=0
    a_center=0.5*R
    a_right=R
    d_delta=1e200 #np.inf
    while d_delta>accuracy:
        delta_predict=_delta_of_a(a_center, R)
        if delta_predict>delta:
            a_left,a_right=a_left,a_center
            a_center=a_left+(a_right-a_left)/2
        elif delta_predict<delta:
            a_left,a_right=a_center,a_right
            a_center=a_left+(a_right-a_left)/2
        else:
            break
        d_delta=abs((delta_predict-delta)/delta)
    return a_center


def get_a(R, delta, accuracy=1e-12):
    """Compute the contact area radius (wrapper)"""
    return _get_a(R, delta, accuracy)


cdef double _delta_of_a(double a, double R):
    cdef double delta
    delta=a/2*np.log((R+a)/(R-a))
    return delta


def delta_of_a(a, R):
    """Compute indentation from contact area radius (wrapper)"""
    return _delta_of_a(a, R)


model_doc = hertz_spherical.__doc__
model_func = hertz_spherical
model_key = "sneddon_spher"
model_name = "spherical indenter (Sneddon)"
parameter_keys = ["E", "R", "nu", "contact_point", "baseline"]
parameter_names = ["Young's Modulus", "Tip Radius",
                   "Poisson's Ratio", "Contact Point", "Force Baseline"]
parameter_units = ["Pa", "m", "", "m", "N"]
valid_axes_x = ["tip position"]
valid_axes_y = ["force"]



"""Test qmap feature"""
import pathlib
import warnings

import numpy as np

from nanite import model, qmap, IndentationGroup


data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map"
jpkfile2 = data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map"


def test_feat_scan_order():
    qm = qmap.QMap(jpkfile)
    order = qm.get_qmap("data: scan order", qmap_only=True)
    assert order[0, 0] == 0
    assert order[0, -1] == 1
    assert order[-1, -1] == 2
    assert order[-1, 0] == 3
    assert np.isnan(order[0, 1])


def test_feat_min_height():
    qm = qmap.QMap(jpkfile)
    qd = qm.get_qmap("data: height base point", qmap_only=True)
    assert np.allclose(qd[0, 0], 40.55030392499141)
    assert np.allclose(qd[0, -1], 47.354988549298945)
    assert np.allclose(qd[-1, -1], 96.1627883099352)
    assert np.allclose(qd[-1, 0], 89.95170867840217)


def test_feat_cp():
    qm = qmap.QMap(jpkfile2)
    # fit data
    for idnt in qm.group:
        idnt.apply_preprocessing(["compute_tip_position",
                                  "correct_force_offset",
                                  "correct_tip_offset",
                                  ])
        inparams = model.model_sneddon_spherical_approximation \
            .get_parameter_defaults()
        inparams["E"].value = 50
        inparams["R"].value = 37.28e-6 / 2
        inparams["baseline"].vary = False

        # Fit with absolute full range
        idnt.fit_model(model_key="sneddon_spher_approx",
                       params_initial=inparams,
                       range_x=(0, 0),
                       range_type="absolute",
                       x_axis="tip position",
                       y_axis="force",
                       segment="approach",
                       weight_cp=2e-6)

    qd = qm.get_qmap("fit: contact point", qmap_only=True)
    vals = qd.flat[~np.isnan(qd.flat)]
    # These are subject to change when contact point preprocessing
    # changes.
    assert np.allclose(vals[0], -1820.431327293334), "gray matter"
    assert np.allclose(vals[2], -3581.010097829371), "white matter"
    assert np.allclose(vals[1], -2106.9835957851064), "background"


def test_feat_emod_nofit():
    qm = qmap.QMap(jpkfile)
    with warnings.catch_warnings(record=True) as w:
        # No data available, because there is no fit
        qd = qm.get_qmap("fit: Young's modulus", qmap_only=True)
        assert len(w) == 4
        assert w[0].category is qmap.DataMissingWarning
    assert np.alltrue(np.isnan(qd))


def test_feat_emod_withfit():
    qm = qmap.QMap(jpkfile2)
    # fit data
    for idnt in qm.group:
        idnt.apply_preprocessing(["compute_tip_position",
                                  "correct_force_offset",
                                  "correct_tip_offset",
                                  ])
        inparams = model.model_sneddon_spherical_approximation \
            .get_parameter_defaults()
        inparams["E"].value = 50
        inparams["R"].value = 37.28e-6 / 2
        inparams["baseline"].vary = False

        # Fit with absolute full range
        idnt.fit_model(model_key="sneddon_spher_approx",
                       params_initial=inparams,
                       range_x=(0, 0),
                       range_type="absolute",
                       x_axis="tip position",
                       y_axis="force",
                       segment="approach",
                       weight_cp=2e-6)

    qd = qm.get_qmap("fit: Young's modulus", qmap_only=True)
    vals = qd.flat[~np.isnan(qd.flat)]
    # gray matter
    assert np.allclose(vals[0], 57.62946472325552, atol=0, rtol=.01)
    # white matter
    assert np.allclose(vals[2], 46.61406865570242, atol=0, rtol=.01)
    # background
    assert np.allclose(vals[1], 17605.034077063443, atol=0, rtol=0.0001)


def test_feat_rating():
    """Reproduces rating in figures 5K-M"""
    qm = qmap.QMap(jpkfile2)
    # fit data
    for idnt in qm.group:
        idnt.apply_preprocessing(["compute_tip_position",
                                  "correct_force_offset",
                                  "correct_tip_offset"])
        inparams = model.model_sneddon_spherical_approximation \
            .get_parameter_defaults()
        inparams["E"].value = 50
        inparams["R"].value = 37.28e-6 / 2
        inparams["baseline"].vary = False

        # Fit with absolute full range
        idnt.fit_model(model_key="sneddon_spher_approx",
                       params_initial=inparams,
                       range_x=(0, 0),
                       range_type="absolute",
                       x_axis="tip position",
                       y_axis="force",
                       segment="approach",
                       weight_cp=2e-6)
        idnt.rate_quality(training_set="zef18",
                          regressor="Extra Trees")

    qd = qm.get_qmap("fit: rating", qmap_only=True)
    vals = qd.flat[~np.isnan(qd.flat)]
    assert np.allclose(vals[0], 9.370435813605962), "gray matter"
    assert np.allclose(vals[2], 4.942804081687071), "white matter"
    assert np.allclose(vals[1], 2.432396277782555), "background"


def test_feat_rating_nofit():
    qm = qmap.QMap(jpkfile)
    with warnings.catch_warnings(record=True) as w:
        # No data available, because there is no fit
        qd = qm.get_qmap("fit: rating", qmap_only=True)
        assert len(w) == 4
        assert w[0].category is qmap.DataMissingWarning
    assert np.alltrue(np.isnan(qd))


def test_get_coords():
    qm = qmap.QMap(jpkfile)

    px = qm.get_coords(which="px")
    refpx = np.array([[0, 0], [9, 0], [9, 9], [0, 9]])
    assert np.all(px == refpx)

    um = qm.get_coords(which="um")
    refum = np.array([[31.972656250000004, -753.5351562500001],
                      [571.8359375000001, -753.90625],
                      [571.8359375000001, -213.73046875000003],
                      [31.855468750000004, -213.73046875000003]])
    assert np.all(um == refum)


def test_get_coords_bad():
    qm = qmap.QMap(jpkfile)
    try:
        qm.get_coords(which="mm")
    except ValueError:
        pass
    else:
        assert False, "Units [mm] should not be supported."


def test_get_qmap():
    qm = qmap.QMap(jpkfile)
    x, y, _ = qm.get_qmap(feature="data: height base point", qmap_only=False)
    assert x.size == 10
    assert y.size == 10


def test_init_with_dataset():
    ds = IndentationGroup(jpkfile)
    qm = qmap.QMap(ds)
    assert qm.shape == (10, 10)


def test_metadata():
    qm = qmap.QMap(jpkfile)
    assert np.allclose(qm.extent,
                       [1.97265625, 601.97265625,
                        -783.53515625, -183.53515625000006])
    assert qm.shape == (10, 10)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import tempfile

from nanite.cli import plotting
from nanite import load_group

data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_plotting():
    # use temporary file
    _, name = tempfile.mkstemp(suffix=".png", prefix="test_nanite_plotting_")
    name = pathlib.Path(name)

    grp = load_group(jpkfile)
    idnt = grp[0]
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset"])
    idnt.fit_model(model_key="hertz_para",
                   params_initial=None,
                   x_axis="tip position",
                   y_axis="force",
                   weight_cp=False,
                   segment="retract")
    plotting.plot_data(idnt=idnt, path=name)
    assert name.stat().st_size > 90000


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Save and load user-rated datasets"""
from functools import lru_cache
import hashlib
import json
import pathlib
import shutil
import tempfile
import time

import h5py
import lmfit
import numpy as np
from sklearn import model_selection

from .._version import version as nanite_version
from ..group import IndentationGroup
from . import rater


class RateManager:
    def __init__(self, path, verbose=0):
        """Manage user-defined rates"""
        #: Path to the manual ratings (directory or .h5 file)
        self.path = pathlib.Path(path)
        #: verbosity level
        self.verbose = verbose
        self._ratings = None

    @staticmethod
    def _get_samples(path, verbose=0):
        rm = RateManager(path, verbose=verbose)
        samples = []
        idr = rater.IndentationRater
        for ds in rm.datasets:
            assert "success" in ds.fit_properties
            features = idr.compute_features(ds)
            samples.append(features)
        return np.array(samples, dtype=float)

    @property
    def datasets(self):
        return [r["data_set"] for r in self.ratings]

    @property
    def ratings(self):
        if self._ratings is None:
            self._ratings = load(self.path, verbose=self.verbose)
        return self._ratings

    @property
    def samples(self):
        """The individual sample ratings computed by afmlib"""
        return RateManager._get_samples(self.path, verbose=self.verbose)

    def export_training_set(self, path):
        path = pathlib.Path(path)
        path.mkdir(parents=True, exist_ok=True)
        raters = rater.IndentationRater.get_feature_funcs()
        samples = self.samples
        for ii, rti in enumerate(raters):
            rpath = path / "train_{}.txt".format(rti[0])
            np.savetxt(rpath, samples[:, ii].flatten(), fmt="%.2e")
        user = self.get_rates(which="user")
        upath = str(path / "train_response.txt")
        np.savetxt(upath, user.flatten(), fmt="%.2e")

    def get_cross_validation_score(self, regressor, training_set=None,
                                   n_splits=20, random_state=42):
        """Regressor cross-validation scoring

        Cross-validation is used to identify regressors that
        over-fit the train set by splitting the train set into
        multiple learn/test sets and quantifying the regressor
        performance for each split.

        Parameters
        ----------
        regressor: str or RegressorMixin
            If a string, must be in `reg_names`.
        training_set: X, y
            If given, do not use self.samples

        Notes
        -----
        A :class:`skimage.model_selection.KFold` cross validator is used
        in combination with the mean squared error score.

        Cross-validation score is computed from samples that are filtered
        with the binary features and only from samples that do not contain
        any nan values.
        """
        ir = rater.get_rater(regressor)

        if training_set:
            X, y = training_set
        else:
            # remove binary features and nans otherwise cross-validation
            # will not work
            X, y = self.get_training_set(prefilter_binary=True,
                                         remove_nans=True)

        # The score is maximized, therefore it is "neg_"
        scoring = 'neg_mean_squared_error'
        loo = model_selection.KFold(n_splits=n_splits, shuffle=True,
                                    random_state=random_state)
        scores = model_selection.cross_val_score(ir.pipeline, X, y,
                                                 scoring=scoring, cv=loo)
        return -scores

    def get_rates(self, which="user", training_set="zef18"):
        """
        which: str
            Which rating to return: "user" or a regressor name
        """
        if which == "user":
            rtngs = np.array([ri["rating"]
                              for ri in load(self.path, meta_only=True)])
        else:
            rt = rater.get_rater(regressor=which, training_set=training_set)
            rtngs = rt.rate(self.samples)
        return rtngs

    def get_training_set(self, which_type="all", prefilter_binary=False,
                         remove_nans=False, transform=False):
        """Return (X, y) training set"""
        X = self.samples
        y = self.get_rates(which="user")

        if prefilter_binary:
            # remove samples with `False` in a binary feature
            ir = rater.IndentationRater(regressor=None)
            bnames, bind = ir.get_feature_names(which_type="binary",
                                                ret_indices=True)
            if bnames:
                X_bool = X[:, bind]

                X_f = []
                y_f = []
                for ii in range(len(y)):
                    if ir._pre_rate(X_bool[ii]):
                        X_f.append(X[ii])
                        y_f.append(y[ii])
                X, y = np.array(X_f), np.array(y_f)

        # must come after prefilter_binary
        if which_type != "all":
            _unames, indices = rater.IndentationRater.get_feature_names(
                which_type=which_type,
                ret_indices=True)
            # remove unwanted features
            if X.size:
                X = X[:, list(indices)]

        if remove_nans:
            X_n = []
            y_n = []
            for ii in range(len(y)):
                if not np.sum(np.isnan(X[ii])):
                    X_n.append(X[ii])
                    y_n.append(y[ii])
            X, y = np.array(X_n), np.array(y_n)

        if transform:
            # Transform data
            ir = rater.IndentationRater(regressor=None,
                                        training_set=(X, y)
                                        )
            X = ir.pipeline.transform(X)
        return X, y


@lru_cache(maxsize=100)
def hash_file(path, blocksize=65536):
    """Compute sha256 hex-hash of a file

    Parameters
    ----------
    path: str or pathlib.Path
        path to the file
    blocksize: int
        block size read from the file

    Returns
    -------
    hex: str
        The first six characters of the hash
    """
    fname = pathlib.Path(path)
    hasher = hashlib.sha256()
    with fname.open('rb') as fd:
        buf = fd.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
    return hasher.hexdigest()[:6]


def load(path, meta_only=False, verbose=0):
    """

    Notes
    -----
    The .fit_properties attribute of each Indentation instance
    is overridden by a simple dictionary,
    so its functionalities are not available anymore.
    """
    path = pathlib.Path(path)
    ratings = []
    if path.is_dir():
        if verbose:
            print("Performing iterative folder search.")
        for fpath in sorted(path.rglob("*.h5")):
            ratings += load(fpath, meta_only=meta_only)
    elif path.suffix == ".h5":
        if verbose:
            print("Loading file '{}'... ".format(path), end="", flush=True)
        ratings += load_hdf5(path, meta_only=meta_only)
        if verbose:
            print("Done.")
    return ratings


def load_hdf5(path, meta_only=False):
    ratings = []
    path = pathlib.Path(path)
    # temporary directory for original data
    tdir = tempfile.mkdtemp(prefix="nanite_rate_data_")
    with h5py.File(path, mode="r") as h5:
        if not meta_only:
            # extract experimental data
            dataset_dict = {}
            for dkey in h5["data"]:
                dset = h5["data"][dkey]
                dbin = dset[...]
                name = dkey + "_" + pathlib.Path(dset.attrs["path"]).name
                dpath = pathlib.Path(tdir) / name
                dbin.tofile(str(dpath))
                dataset_dict[dkey] = IndentationGroup(dpath)
        # load individual curves
        for akey in h5["analysis"]:
            h5gr = h5["analysis"][akey]
            attrs = h5gr.attrs
            if not meta_only:
                indent = dataset_dict[attrs["data hash"]].get_enum(
                    attrs["data enum"])
                indent["fit"] = h5gr["fit"][...]
                indent["fit range"] = h5gr["fit range"][...]
                indent["force"] = h5gr["force"][...]
                indent["fit residuals"] = h5gr["fit residuals"][...]
                indent["tip position"] = h5gr["tip position"][...]
                indent["segment"] = h5gr["segment"][...]
            fit_properties = {}
            fkeys = [key for key in attrs if key.startswith("fit ")]
            for fkey in fkeys:
                key = fkey[4:]
                val = attrs[fkey]
                if key.startswith("params"):
                    parms = lmfit.Parameters()
                    parms.loads(val)
                    val = parms
                elif key == "preprocessing":
                    val = val.split(",")
                elif key in ["preprocessing_options", "method_kws"]:
                    val = json.loads(val)
                elif key == "range_x":
                    val = val.strip("[]() ").split(",")
                    val = (float(val[0]), float(val[1]))
                fit_properties[key] = val
            rating = {
                "name": attrs["user name"],
                "rating": attrs["user rate"],
                "comment": attrs["user comment"],
                "enum": attrs["data enum"],
                "fit properties": fit_properties,
            }
            if not meta_only:
                # This overrides the FitProperties class!
                indent.fit_properties = fit_properties
                rating["data_set"] = indent
            ratings.append(rating)

    shutil.rmtree(tdir, ignore_errors=True)
    return ratings


def save_hdf5(h5path, indent, user_rate, user_name, user_comment, h5mode="a"):
    """Store all relevant data of a user rating into an hdf5 file

    Parameters
    ----------
    h5path: str or pathlib.Path
        Path to HDF5 rating container where data will be stored
    indent: nanite.Indentation
        The experimental data processed and fitted with nanite
    user_rate: float
        Rating given by the user
    user_name: str
        Name of the rating user
    """
    dkw = {"fletcher32": True,
           "compression": "gzip",
           "compression_opts": 9}
    with h5py.File(h5path, mode=h5mode) as h5:
        # store raw experimental data as binary array
        data = h5.require_group("data")
        dhash = hash_file(indent.path)
        if dhash not in data:
            meas = data.create_dataset(
                dhash,
                data=np.fromfile(str(indent.path), dtype=bool),
                **dkw
            )
            meas.attrs["path"] = str(indent.path)
        # store indentation data along with the user rate
        ana = h5.require_group("analysis")
        idd = "{}_{}".format(dhash, indent.enum)
        if idd in ana:
            # Only allow overriding of user data if fit matches.
            # Otherwise, the rating might be wrong.
            if not np.allclose(indent["fit"], ana[idd]["fit"], equal_nan=True):
                raise ValueError("Cannot store rating for different fit in "
                                 "same rating container!")
            out = ana[idd]
        else:
            out = ana.create_group(idd)
            out.attrs["data enum"] = indent.enum
            out.attrs["data hash"] = dhash
            for key in indent.fit_properties:
                val = indent.fit_properties[key]
                if key.startswith("params_"):
                    val = val.dumps()
                elif key == "preprocessing":
                    val = ",".join(val)
                elif key in ["preprocessing_options", "method_kws"]:
                    val = json.dumps(val)
                elif key == "range_x":
                    val = str(val)
                out.attrs["fit {}".format(key)] = val

            out.create_dataset("fit",
                               data=indent["fit"][...],
                               **dkw)
            out.create_dataset("fit range",
                               data=indent["fit range"][...],
                               **dkw)
            out.create_dataset("force",
                               data=indent["force"][...],
                               **dkw)
            out.create_dataset("fit residuals",
                               data=indent["fit residuals"][...],
                               **dkw)
            out.create_dataset("tip position",
                               data=indent["tip position"][...],
                               **dkw)
            out.create_dataset("segment",
                               data=indent["segment"][...],
                               **dkw)
        # update user data in any case
        out.attrs["user comment"] = user_comment
        out.attrs["user name"] = user_name
        out.attrs["user rate"] = user_rate
        out.attrs["user time"] = time.time()
        out.attrs["user time str"] = time.ctime()
        # add library versions for debugging
        out.attrs["nanite version"] = nanite_version
        out.attrs["h5py version"] = h5py.__version__


def hdf5_rated(h5path, indent):
    """Test whether an indentation has already been rated

    Returns
    -------
    is_rated, rating, comment
    """
    is_rated = False
    rating = -1
    comment = ""
    h5path = pathlib.Path(h5path)
    if h5path.exists():
        with h5py.File(h5path, mode="r") as h5:
            if "analysis" in h5:
                ana = h5["analysis"]
                dhash = hash_file(indent.path)
                idd = "{}_{}".format(dhash, indent.enum)
                if idd in ana:
                    is_rated = True
                    rating = ana[idd].attrs["user rate"]
                    comment = ana[idd].attrs["user comment"]
    return is_rated, rating, comment



import warnings

from .residuals import compute_contact_point_weights  # noqa: F401


def weight_cp(*args, **kwargs):
    warnings.warn(
        "The 'weight' module is deprecated. Please use "
        "'from nanite.model.residuals import compute_contact_point_weights'!",
        DeprecationWarning)
    return compute_contact_point_weights(*args, **kwargs)



"""Test point of contact (POC) estimation"""
import pathlib

import numpy as np
import pytest

from nanite import poc, IndentationGroup


data_path = pathlib.Path(__file__).resolve().parent / "data"


@pytest.mark.parametrize("method,contact_point", [
    ["gradient_zero_crossing", 1895],
    ["fit_constant_line", 1919],
    ["fit_constant_polynomial", 1898],
    ["fit_line_polynomial", 1899],
    ["frechet_direct_path", 1903],
    ["deviation_from_baseline", 1908],
])
def test_poc_estimation(method, contact_point):
    fd = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset"])
    assert poc.compute_poc(fd["force"], method) == contact_point


@pytest.mark.parametrize("method,contact_point", [
    ["gradient_zero_crossing", 1895],
    ["fit_constant_line", 1919],
    ["fit_constant_polynomial", 1898],
    ["fit_line_polynomial", 1899],
    ["frechet_direct_path", 1903],
    ["deviation_from_baseline", 1908],
])
def test_poc_estimation_details(method, contact_point):
    fd = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset"])
    _, details = poc.compute_poc(fd["force"], method, ret_details=True)
    assert np.all(np.array(details["plot poc"][0]) == contact_point)


@pytest.mark.parametrize("method,contact_point", [
    ["gradient_zero_crossing", 1895],
    ["fit_constant_line", 1919],
    ["fit_constant_polynomial", 1898],
    ["fit_line_polynomial", 1899],
    ["frechet_direct_path", 1903],
    ["deviation_from_baseline", 1908],
])
def test_poc_estimation_via_indent(method, contact_point):
    fd = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"],
                           options={"correct_tip_offset": {"method": method}})
    assert np.argmin(np.abs(fd["tip position"])) == contact_point


def test_poc_details_deviation_from_baseline():
    fd = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    details = fd.apply_preprocessing(
        ["compute_tip_position", "correct_force_offset", "correct_tip_offset"],
        options={"correct_tip_offset": {"method": "deviation_from_baseline"}},
        ret_details=True)
    pocd = details["correct_tip_offset"]
    for key in ["plot force",
                "plot baseline mean",
                "plot baseline threshold",
                "plot poc"]:
        assert key in pocd
    assert np.allclose(
        np.mean(pocd["plot baseline threshold"][1]),
        8.431890003513514e-11,
        atol=0)
    assert np.allclose(
        np.mean(pocd["plot baseline mean"][1]),
        -7.573822405258677e-12,
        atol=0)


def test_poc_details_fit_line_polynomial():
    fd = IndentationGroup(
        data_path
        / "fmt-jpk-fd_single_tilted-baseline-mitotic_2021-01-29.jpk-force")[0]
    details = fd.apply_preprocessing(
        ["compute_tip_position", "correct_tip_offset"],
        options={"correct_tip_offset": {"method": "fit_line_polynomial"}},
        ret_details=True)
    pocd = details["correct_tip_offset"]
    for key in ["plot force",
                "plot fit",
                "plot poc"]:
        assert key in pocd
    assert np.allclose(
        pocd["plot poc"][0][0],
        15602,
        atol=0)
    assert np.allclose(
        fd["force"][15602],
        1.7313584441928315e-09,
        atol=0,
    )


def test_poc_frechet_direct_path():
    force = np.zeros(100)
    force[50:] = np.arange(50)**2
    cp = poc.poc_frechet_direct_path(force)
    assert cp == 62  # makes sense



"""Test rating features"""
import pathlib

import numpy as np

from nanite import IndentationGroup, model
from nanite.rate import features


data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def setup_indent():
    idnt = IndentationGroup(jpkfile)[0]
    idnt.apply_preprocessing(["compute_tip_position"])
    inparams = model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8e-5)

    # Fit with absolute full range
    idnt.fit_model(model_key="hertz_para",
                   params_initial=inparams,
                   range_x=(0, 0),
                   range_type="absolute",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)
    return idnt


def test_basic_nofit():
    idnt = IndentationGroup(jpkfile)[0]
    feat = features.IndentationFeatures(idnt)
    assert not feat.is_fitted
    assert not feat.is_valid
    assert not feat.has_contact_point
    try:
        feat.contact_point
    except ValueError:
        pass
    else:
        assert False, "without fit - should not have a CP"


def test_basic_withfit():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    assert feat.is_fitted
    assert feat.is_valid
    assert feat.has_contact_point

    cp = feat.contact_point
    # This will change if the the default method of CP detection
    # changes.
    assert np.allclose(cp, 1.8029310201193193e-05)
    # Simple check for meta
    assert feat.meta["point count"] == 4000


def test_get_features():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    samples, names = feat.compute_features(idnt, ret_names=True)

    # These are the samples from the original release
    ref = {
        'feat_bin_apr_spikes_count': 1.0,
        'feat_bin_cp_position': 1.0,
        'feat_bin_size': 1.0,
        'feat_con_apr_flatness': 0.4157949790794979,
        'feat_con_apr_size': 0.04400000000000004,
        'feat_con_apr_sum': 0.3293102976667975,
        'feat_con_bln_slope': 0.6287923407474181,
        'feat_con_bln_variation': 0.46602504647335763,
        'feat_con_cp_curvature': -0.15481927352155514,
        'feat_con_cp_magnitude': 0.005076769161668967,
        'feat_con_idt_maxima_75perc': 0.3184029609825442,
        'feat_con_idt_monotony': 0.0,
        'feat_con_idt_spike_area': 0.4262935122879405,
        'feat_con_idt_sum': 0.19263421098385752,
        'feat_con_idt_sum_75perc': 0.015026423315718608
    }
    for key in ref:
        idx = names.index(key)
        assert np.allclose(samples[idx], ref[key]), "Mismatch '{}'".format(key)


def test_get_feature_funcs_order():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    funcs = feat.get_feature_funcs()
    samples = feat.compute_features(idnt)
    for (_, func), samp in zip(funcs, samples):
        assert np.allclose(func(feat), samp)


def test_get_feature_names_bad_names():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    try:
        feat.get_feature_names(names=["feat_con_apr_flatness",
                                      "feat_con_unknown"])
    except ValueError:
        pass
    else:
        assert False, "Unknown names should not work"


def test_get_feature_names_bad_type():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    try:
        feat.get_feature_names(which_type="unknown")
    except ValueError:
        pass
    else:
        assert False, "Unknown types should not work"


def test_get_feature_names_type():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    # binary
    names = feat.get_feature_names(which_type="binary")
    for nn in names:
        assert nn.startswith("feat_bin_")
    # continuous
    names = feat.get_feature_names(which_type="continuous")
    for nn in names:
        assert nn.startswith("feat_con_")


def test_get_feature_names_mix():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    # binary
    names = feat.get_feature_names(which_type="binary",
                                   names=["feat_bin_size",
                                          "feat_con_idt_sum"])
    assert len(names) == 1
    assert names[0] == "feat_bin_size"


def test_get_feature_names_indices():
    idnt = setup_indent()
    feat = features.IndentationFeatures(idnt)
    # binary
    names, idx = feat.get_feature_names(which_type="binary",
                                        names=["feat_bin_size",
                                               "feat_con_idt_sum"],
                                        ret_indices=True)
    nall = feat.get_feature_names()
    assert idx[0] == nall.index(names[0])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for _key in list(loc.keys()):
        if _key.startswith("test_") and hasattr(loc[_key], "__call__"):
            loc[_key]()



"""Test of inverse function"""
import numpy as np

from nanite.model import model_sneddon_spherical as hertz_spherical


def test_get_a_sneddon():
    r = 40e-6
    delta_real = [0.1e-6, 2e-6, 5e-6]
    delta_predict = np.zeros_like(delta_real)
    for i in range(len(delta_real)):
        a = hertz_spherical.get_a(r, delta_real[i], accuracy=1e-09)
        delta_predict[i] = hertz_spherical.delta_of_a(a, r)

    assert np.allclose(delta_real, delta_predict, rtol=1e-06, atol=0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test model integrity"""
from nanite import model


NAME_MAPPING = {
    "alpha": ["Face Angle", "Half Cone Angle"],
    "baseline": ["Force Baseline"],
    "contact_point": ["Contact Point"],
    "E": ["Young's Modulus"],
    "nu": ["Poisson's Ratio"],
    "R": "Tip Radius",
}

UNIT_MAPPING = {
    "alpha": "°",
    "baseline": "N",
    "contact_point": "m",
    "E": "Pa",
    "nu": "",
    "R": "m",
}


def test_model_parameter_name_order():
    for key in model.models_available:
        md = model.models_available[key]
        std = md.get_parameter_defaults()
        for n1, n2 in zip(list(std.keys()), md.parameter_keys):
            assert n1 == n2, "Parameter defaults probably in wrong order!"


def test_model_parameter_names():
    for key in model.models_available:
        md = model.models_available[key]
        for key2, nn in zip(md.parameter_keys, md.parameter_names):
            if key2 in NAME_MAPPING:
                assert nn in NAME_MAPPING[key2], "bad {} in {}".format(
                    key2, md)
            else:
                msg = "Parameter {} not registered for test!".format(key2)
                assert False, msg


def test_model_parameter_units():
    for key in model.models_available:
        md = model.models_available[key]
        for key2, un in zip(md.parameter_keys, md.parameter_units):
            if key2 in UNIT_MAPPING:
                assert UNIT_MAPPING[key2] == un, "bad {} in {}".format(
                    key2, md)
            else:
                msg = "Parameter {} not registered for test!".format(key2)
                assert False, msg


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for _key in list(loc.keys()):
        if _key.startswith("test_") and hasattr(loc[_key], "__call__"):
            loc[_key]()



"""scikit-learn regressors and their keyword arguments"""
from sklearn import ensemble, svm, tree


reg_dict = {
    "AdaBoost": [
        ensemble.AdaBoostRegressor,
        {"learning_rate": .18,
         "n_estimators": 100,
         "random_state": 42,
         },
    ],
    "Decision Tree": [
        tree.DecisionTreeRegressor,
        {"max_depth": 6,
         "min_samples_leaf": 14,
         "min_samples_split": 3,
         "random_state": 42,
         }
    ],
    "Extra Trees": [
        ensemble.ExtraTreesRegressor,
        {"max_depth": 15,
         "min_samples_leaf": 1,
         "min_samples_split": 10,
         "n_estimators": 100,
         "random_state": 42,
         }
    ],
    "Gradient Tree Boosting": [
        ensemble.GradientBoostingRegressor,
        {"max_depth": 4,
         "min_samples_leaf": 4,
         "min_samples_split": 7,
         "n_estimators": 100,
         "random_state": 42,
         }
    ],
    "Random Forest": [
        ensemble.RandomForestRegressor,
        {"max_depth": 15,
         "min_samples_leaf": 2,
         "min_samples_split": 6,
         "n_estimators": 100,
         "random_state": 42,
         }
    ],
    "SVR (linear kernel)": [
        svm.LinearSVR,
        {"C": 2,
         "epsilon": .75,
         "random_state": 42,
         },
    ],
    "SVR (RBF kernel)": [
        svm.SVR,
        {"kernel": "rbf",
         "C": 25,
         "epsilon": 0.75,
         },
    ],
}


#: List of available default regressor names
reg_names = sorted(reg_dict.keys())

#: List of tree-based regressor class names (used for keyword defaults in
#: :class:`IndentationRater`)
reg_trees = ["AdaBoostRegressor",
             "DecisionTreeRegressor",
             "ExtraTreesRegressor",
             "GradientBoostingRegressor",
             "RandomForestRegressor"]



"""Test preprocessing"""
import pathlib

import numpy as np
import pytest

from nanite import IndentationGroup
from nanite import preproc


data_path = pathlib.Path(__file__).resolve().parent / "data"
bad_files = list(data_path.glob("bad*"))


def test_autosort():
    unsorted = ["correct_force_offset",
                "correct_tip_offset",
                "compute_tip_position"]
    expected = ["correct_force_offset",
                "compute_tip_position",
                "correct_tip_offset"]
    actual = preproc.autosort(unsorted)
    assert expected == actual


def test_autosort2():
    unsorted = ["correct_split_approach_retract",
                "correct_force_offset",
                "correct_tip_offset",
                "compute_tip_position",
                ]
    expected = ["compute_tip_position",
                "correct_split_approach_retract",
                "correct_force_offset",
                "correct_tip_offset",
                ]
    actual = preproc.autosort(unsorted)
    assert (actual.index("correct_split_approach_retract")
            > actual.index("compute_tip_position"))
    assert (actual.index("correct_tip_offset")
            > actual.index("compute_tip_position"))
    assert expected == actual


def test_autosort3():
    unsorted = ["smooth_height",
                "correct_split_approach_retract",
                "correct_force_offset",
                "correct_tip_offset",
                "compute_tip_position",
                ]
    expected = ["compute_tip_position",
                "correct_split_approach_retract",
                "smooth_height",
                "correct_force_offset",
                "correct_tip_offset",
                ]
    actual = preproc.autosort(unsorted)
    assert expected == actual
    assert (actual.index("correct_split_approach_retract")
            > actual.index("compute_tip_position"))
    assert (actual.index("correct_tip_offset")
            > actual.index("compute_tip_position"))
    assert (actual.index("smooth_height")
            > actual.index("correct_split_approach_retract"))


def test_check_order():
    with pytest.raises(ValueError, match="Wrong optional step order"):
        preproc.check_order([
            "smooth_height",
            "correct_split_approach_retract"])


def test_correct_split_approach_retract():
    fd = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]

    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"])
    assert fd.appr["segment"].size == 2000
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset",
                            "correct_split_approach_retract"])
    assert fd.appr["segment"].size == 2006


def test_get_steps_required():
    req_act = preproc.get_steps_required("correct_tip_offset")
    req_exp = ["compute_tip_position"]
    assert req_act == req_exp


@pytest.mark.filterwarnings('ignore::nanite.preproc.CannotSplitWarning',
                            'ignore::UserWarning')
def test_process_bad():
    for bf in bad_files:
        ds = IndentationGroup(bf)
        for idnt in ds:
            # Walk through the standard analysis pipeline without
            # throwing any exceptions.
            idnt.apply_preprocessing(["compute_tip_position",
                                      "correct_force_offset",
                                      "correct_tip_offset",
                                      "correct_split_approach_retract"])
            idnt.fit_model()


@pytest.mark.filterwarnings(
    'ignore::nanite.smooth.DoubledSmoothingWindowWarning:')
def test_smooth():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset",
                              "correct_tip_offset"])
    orig = np.array(idnt["tip position"], copy=True)
    # now apply smoothing filter
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset",
                              "correct_tip_offset",
                              "smooth_height"])
    new = np.array(idnt["tip position"], copy=True)
    assert not np.all(orig == new)


def test_unknown_method():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    with pytest.raises(KeyError, match="unknown_method"):
        idnt.apply_preprocessing(["compute_tip_position",
                                  "unknown_method"])


def test_wrong_order():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    with pytest.raises(ValueError, match="requires the steps"):
        # order matters
        idnt.apply_preprocessing(["correct_tip_offset",
                                  "compute_tip_position"])


def test_wrong_order_2():
    idnt = IndentationGroup(data_path / "fmt-jpk-fd_spot3-0192.jpk-force")[0]
    with pytest.raises(ValueError, match="requires the steps"):
        # order matters
        idnt.apply_preprocessing(["correct_split_approach_retract",
                                  "compute_tip_position"])


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import matplotlib.pylab as plt
import numpy as np


def plot_data(idnt, figure=None, add_text="", path=None):
    if figure is None:
        figure = plt.figure(figsize=(10, 7))
    else:
        figure.clear()
    ax = figure.add_axes([.07, .1, .90, .87])
    ax.set_xlabel("tip position [µm]")
    ax.set_ylabel("force [nN]")
    # overview plot
    ax.plot(idnt["tip position"] * 1e6, idnt["fit"] * 1e9)
    ax.plot(idnt["tip position"] * 1e6, idnt["force"] * 1e9)
    # inset with indentation
    axin = figure.add_axes([.3, .3, .65, .65])
    axin.patch.set_alpha(0.5)
    # inset limits depend on contact point
    axin.plot(idnt["tip position"] * 1e6, idnt["fit"] * 1e9, label="fit")
    axin.plot(idnt["tip position"] * 1e6, idnt["force"] * 1e9, label="data")
    axin.legend(loc="upper right")
    axin.grid()
    cp = idnt.fit_properties["params_fitted"]["contact_point"].value * 1e6
    xmin = idnt["tip position"].min() * 1e6
    dx = np.abs(cp - xmin)
    ymin = idnt["force"][idnt["segment"] == 0].min() * 1e9
    ymax = idnt["force"][idnt["segment"] == 0].max() * 1e9
    dy = ymax - ymin
    axin.set_xlim(xmin - dx/7, cp + dx/2)
    axin.set_ylim(ymin - dy/7, ymax + dy/7)
    axin.set_xticks([])
    axin.set_ylabel("close-up")
    # inset with residuals
    axin2 = figure.add_axes([.3, .15, .65, .14])
    axin2.patch.set_alpha(0.5)
    axin2.plot(idnt["tip position"] * 1e6,
               idnt["fit residuals"] * 1e9, label="fit")
    axin2.set_xlim(xmin - dx/7, cp + dx/2)
    axin2.set_ylabel("fit residuals")
    axin2.axhline(0, color="gray")
    axin2.set_ylim(-dy/10, dy/10)
    # display fitted parameters
    text = "Fit parameters:\n"
    text += "model: {}\n".format(idnt.fit_properties["model_key"])
    params = idnt.fit_properties["params_fitted"]
    for p in params:
        text += "{}={:.2e}\n".format(p, params[p].value)
    text += "\n\n" + add_text
    axin.text(cp, ymax, text, horizontalalignment="right",
              verticalalignment="top")

    if path is not None:
        figure.savefig(path)
        plt.close()



"""Test opening of file format variations"""
import pathlib
import numpy as np

from nanite import IndentationGroup

data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_process_flipsign():
    # This is a curve extracted from a map file. When loading it
    # with nanite, the sign of the force curve was flipped.
    flipped = (data_path
               / "fmt-jpk-fd_flipsign_2015.05.22-15.31.49.352.jpk-force")
    idnt = IndentationGroup(flipped)[0]
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset",
                              "correct_tip_offset",
                              "correct_split_approach_retract"])
    # We set the baseline fixed, because this test was written so)
    params_initial = idnt.get_initial_fit_parameters(model_key="hertz_para")
    params_initial["baseline"].set(vary=False)
    idnt.fit_model(model_key="hertz_para", weight_cp=False,
                   params_initial=params_initial)
    assert np.allclose(idnt.fit_properties["params_fitted"]["E"].value,
                       5230.008779761989,
                       atol=1,
                       rtol=0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test basic fitting"""
import pathlib

import numpy as np

import nanite
from nanite import IndentationGroup
import pytest


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_lmfit_method():
    ds1 = IndentationGroup(jpkfile)
    apret = ds1[0]
    apret.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8321e-5)

    # Fit with absolute full range
    kwargs = dict(model_key="hertz_para",
                  params_initial=inparams,
                  range_x=(0, 0),
                  range_type="absolute",
                  x_axis="tip position",
                  y_axis="force",
                  segment="approach",
                  weight_cp=False)

    apret.fit_model(**kwargs)
    params1 = apret.fit_properties["params_fitted"]
    assert np.allclose(params1["contact_point"].value,
                       1.802931023582261e-05,
                       rtol=0,
                       atol=0.000000000005,
                       )

    # make sure leastsq is the default
    apret.fit_model(method="leastsq", **kwargs)
    params2 = apret.fit_properties["params_fitted"]
    assert params2["contact_point"].value == params1["contact_point"]

    # use a different method
    apret.fit_model(method="nelder", method_kws={"max_nfev": 5}, **kwargs)
    params3 = apret.fit_properties["params_fitted"]
    assert params3["contact_point"].value != params1["contact_point"]
    assert np.allclose(params3["contact_point"].value,
                       1.8779025e-05,
                       rtol=0,
                       atol=0.000000000005,
                       )


@pytest.mark.parametrize("gcf_k", [0.1, 0.23, 0.3, 1/np.pi, 0.5, 0.6, 1.0])
def test_gcf_k_no_change_in_contact_point(gcf_k):
    """Fit result for contact point does not change with gcf_k"""
    ds1 = IndentationGroup(jpkfile)
    apret = ds1[0]
    apret.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8321e-5)
    # sane initial fit parameters with gcf_k in mind
    inparams["E"].set(inparams["E"].value * (1/gcf_k)**(3/2))

    # Fit with absolute full range
    kwargs = dict(model_key="hertz_para",
                  params_initial=inparams,
                  range_x=(0, 0),
                  range_type="absolute",
                  x_axis="tip position",
                  y_axis="force",
                  segment="approach",
                  weight_cp=False,
                  gcf_k=gcf_k)

    apret.fit_model(**kwargs)
    params1 = apret.fit_properties["params_fitted"]
    assert np.allclose(params1["contact_point"].value,
                       1.802931023582261e-05,
                       rtol=0,
                       atol=0.000000000005,
                       )


@pytest.mark.parametrize("gcf_k", [0.1, 0.23, 0.3, 1/np.pi, 0.5, 0.6, 1.0])
def test_gcf_k_no_change_in_fitted_curve(gcf_k):
    """Fit result for contact point does not change with gcf_k"""
    ds1 = IndentationGroup(jpkfile)
    apret = ds1[0]
    apret.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8321e-5)
    # sane initial fit parameters with gcf_k in mind
    inparams["E"].set(inparams["E"].value * (1/gcf_k)**(3/2))

    # Fit with absolute full range
    kwargs = dict(model_key="hertz_para",
                  params_initial=inparams,
                  range_x=(0, 0),
                  range_type="absolute",
                  x_axis="tip position",
                  y_axis="force",
                  segment="approach",
                  weight_cp=False,
                  gcf_k=gcf_k)

    apret.fit_model(**kwargs)
    assert np.allclose(np.nanmax(apret["fit"]),
                       3.496319691452543e-09,
                       atol=0.000001e-9,
                       rtol=0)
    assert np.allclose(np.nanmax(apret["fit residuals"]),
                       2.233069676202635e-10,
                       atol=0.000001e-10,
                       rtol=0)


@pytest.mark.parametrize("gcf_k", [0.1, 0.23, 0.3, 1/np.pi, 0.5, 0.6, 1.0])
def test_gcf_k_scaling_of_youngs_modulus(gcf_k):
    """Fit result for Young's modulus should scale with gcf_k"""
    ds1 = IndentationGroup(jpkfile)
    apret = ds1[0]
    apret.apply_preprocessing(["compute_tip_position", "correct_tip_offset"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(0)

    # Fit with absolute full range
    kwargs = dict(model_key="hertz_para",
                  params_initial=inparams,
                  range_x=(0, 0),
                  range_type="absolute",
                  x_axis="tip position",
                  y_axis="force",
                  segment="approach",
                  weight_cp=False,
                  gcf_k=gcf_k)

    apret.fit_model(**kwargs)
    params1 = apret.fit_properties["params_fitted"]
    # proportionality between E and gcf_k
    corrval = 14741.950622347102 * (1/gcf_k)**(3/2)
    assert np.allclose(params1["E"].value,
                       corrval,
                       rtol=0.0001,
                       atol=0,
                       )


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib

import nanite


data_dir = pathlib.Path(__file__).parent / "data"


def test_load_model_from_file():
    mpath = data_dir / "model_external_basic.py"
    md = nanite.model.load_model_from_file(mpath, register=True)
    assert md.model_key == "hans_peter"
    assert md.model_key in nanite.model.models_available
    nanite.model.deregister_model(md)
    assert md.model_key not in nanite.model.models_available


def test_load_model_from_model():
    mpath = data_dir / "model_external_basic.py"
    md = nanite.model.load_model_from_file(mpath, register=False)
    assert md.model_key == "hans_peter"
    assert md.model_key not in nanite.model.models_available
    md2 = nanite.model.register_model(md)
    assert md is md2
    assert md.model_key in nanite.model.models_available
    nanite.model.deregister_model(md)
    assert md.model_key not in nanite.model.models_available



import warnings

import afmformats
from afmformats.afm_qmap import qmap_feature
import numpy as np

from .read import get_load_data_modality_kwargs


class DataMissingWarning(UserWarning):
    pass


class QMap(afmformats.AFMQMap):
    def __init__(self, path_or_group, meta_override=None, callback=None):
        """Quantitative force spectroscopy map handling

        Parameters
        ----------
        path_or_group: str or pathlib.Path or afmformats.afm_group.AFMGroup
            The path to the data file or an instance of `AFMGroup`
        meta_override: dict
            Dictionary with metadata that is used when loading the data
            in `path`.
        callback: callable or None
            A method that accepts a float between 0 and 1
            to externally track the process of loading the data.
        """
        super(QMap, self).__init__(
            path_or_group=path_or_group,
            meta_override=meta_override,
            callback=callback,
            **get_load_data_modality_kwargs()
        )

    @staticmethod
    @qmap_feature(name="fit: contact point",
                  unit="nm",
                  cache=False)
    def feat_fit_contact_point(idnt):
        """Contact point of the fit"""
        if idnt.fit_properties.get("success", False):
            # use cached contact point
            params = idnt.fit_properties["params_fitted"]
            value = params["contact_point"].value * 1e9
        else:
            msg = "The experimental data has not been fitted. Please call " \
                  + "`idnt.fit_model` manually for {}!".format(idnt)
            warnings.warn(msg, DataMissingWarning)
            value = np.nan

        return value

    @staticmethod
    @qmap_feature(name="fit: Young's modulus",
                  unit="Pa",
                  cache=False)
    def feat_fit_youngs_modulus(idnt):
        """Young's modulus"""
        if idnt.fit_properties.get("success", False):
            # use cached young's modulus
            value = idnt.fit_properties["params_fitted"]["E"].value
        else:
            msg = "The experimental data has not been fitted. Please call " \
                  + "`idnt.fit_model` manually for {}!".format(idnt)
            warnings.warn(msg, DataMissingWarning)
            value = np.nan

        return value

    @staticmethod
    @qmap_feature(name="fit: rating",
                  unit="",
                  cache=False)
    def feat_meta_rating(idnt):
        """Rating"""
        if idnt._rating is None:
            msg = "The experimental data has not been rated. Please call " \
                  + "`idnt.rate_quality` manually for {}!".format(idnt)
            warnings.warn(msg, DataMissingWarning)
            value = np.nan
        else:
            # use cached rating
            value = idnt._rating[-1]
        return value



import lmfit
import numpy as np
from numpy import pi


def get_parameter_defaults():
    """Return the default model parameters"""
    # The order of the parameters must match the order
    # of ´parameter_names´ and ´parameter_keys´.
    params = lmfit.Parameters()
    params.add("E", value=3e3, min=0)
    params.add("alpha", value=25, min=0, max=90, vary=False)
    params.add("nu", value=.5, min=0, max=0.5, vary=False)
    params.add("contact_point", value=0)
    params.add("baseline", value=0)
    return params


def hertz_conical(delta, E, alpha, nu, contact_point=0, baseline=0):
    r"""Hertz model for a conical indenter

    .. math::

        F = \frac{2\tan\alpha}{\pi}
            \frac{E}{1-\nu^2}
            \delta^2

    Parameters
    ----------
    delta: 1d ndarray
        Indentation [m]
    E: float
        Young's modulus [N/m²]
    alpha: float
        Half cone angle [degrees]
    nu: float
        Poisson's ratio
    contact_point: float
        Indentation offset [m]
    baseline: float
        Force offset [N]

    Returns
    -------
    F: float
        Force [N]

    Notes
    -----
    These approximations are made by the Hertz model:

    - The sample is isotropic.
    - The sample is a linear elastic solid.
    - The sample is extended infinitely in one half space.
    - The indenter is not deformable.
    - There are no additional interactions between sample and indenter.

    Additional assumptions:

    - infinitely sharp probe

    References
    ==========
    Love (1939) :cite:`Love1939`,
    Sneddon (1965) :cite:`Sneddon1965` (equation 6.4)
    """
    aa = 2*np.tan(alpha*pi/180)/pi * E/(1-nu**2)
    root = contact_point-delta
    pos = root > 0
    bb = np.zeros_like(delta)
    bb[pos] = root[pos]**2
    return aa*bb + baseline


model_doc = hertz_conical.__doc__
model_func = hertz_conical
model_key = "hertz_cone"
model_name = "conical indenter (Hertz)"
parameter_keys = ["E", "alpha", "nu", "contact_point", "baseline"]
parameter_names = ["Young's Modulus", "Half Cone Angle",
                   "Poisson's Ratio", "Contact Point", "Force Baseline"]
parameter_units = ["Pa", "°", "", "m", "N"]
valid_axes_x = ["tip position"]
valid_axes_y = ["force"]



import lmfit
import numpy as np


def get_parameter_defaults():
    """Return the default model parameters"""
    # The order of the parameters must match the order
    # of ´parameter_names´ and ´parameter_keys´.
    params = lmfit.Parameters()
    params.add("E", value=3e3, min=0)
    params.add("R", value=10e-6, min=0, vary=False)
    params.add("nu", value=.5, min=0, max=0.5, vary=False)
    params.add("contact_point", value=0)
    params.add("baseline", value=0)
    return params


def hertz_paraboloidal(delta, E, R, nu, contact_point=0, baseline=0):
    """This is identical to the Hertz parabolic indenter model"""
    aa = 4/3 * E/(1-nu**2)*np.sqrt(R)
    root = contact_point-delta
    pos = root > 0
    bb = np.zeros_like(delta)
    bb[pos] = (root[pos])**(3/2)
    return aa*bb + baseline


model_doc = hertz_paraboloidal.__doc__
model_func = hertz_paraboloidal
model_key = "hans_peter"
model_name = "Hans Peter's model"
parameter_keys = ["E", "R", "nu", "contact_point", "baseline"]
parameter_names = ["Young's Modulus", "Tip Radius",
                   "Poisson's Ratio", "Contact Point", "Force Baseline"]
parameter_units = ["Pa", "m", "", "m", "N"]
valid_axes_x = ["tip position"]
valid_axes_y = ["force"]



"""smooth data"""
import warnings

import numpy as np
import scipy.ndimage as im


class DoubledSmoothingWindowWarning(UserWarning):
    pass


def smooth_axis(data, window=15):
    """
    Smooth a 1D  data array with a median filter of `width`.

    See Also
    --------
    smooth_axis_monotone
    """
    smooth = im.median_filter(data, size=(window,), mode="nearest")
    return smooth


def smooth_axis_monotone(data, window=15, max_iter=1000):
    """
    Smooth a 1D  data array with a median filter of `width`.
    This method makes sure that the data is monotonously
    increasing or decreasing, by increasing the window
    size automatically. If this happens, a
    :class:`DoubledSmoothingWindowWarning` is raised.

    Window should be an odd number.

    See Also
    --------
    smooth_axis
    """
    smooth = smooth_axis(data, window=window)
    gradient = np.gradient(smooth)

    for _ in range(max_iter):
        if np.abs(np.sum(gradient)) == np.sum(np.abs(gradient)):
            break
        window = window * 2 + 1
        smooth = smooth_axis(data, window=window)
        gradient = np.gradient(smooth)
        warnings.warn("Automatically doubled smoothing `window` size to "
                      + "{}. You might consider using a ".format(window)
                      + "larger value by default.",
                      DoubledSmoothingWindowWarning)
    else:
        raise ValueError("Reached `max_iter`={}".format(max_iter))

    for _ in range(max_iter):
        if np.unique(smooth).size == smooth.size:
            break
        # Keep axis monotonous.
        # get the first element with equal values
        equal = []
        myset = False
        for ii in range(smooth.size):
            if smooth[ii+1] == smooth[ii]:
                equal.append(ii+1)
                myset = True
            elif myset is True:
                # continue with these values in the next for-loop
                break
            if smooth.size == ii+2:
                # abort
                break

        for count, idx in enumerate(equal):
            try:
                smooth[idx] += (smooth[equal[-1]+1] -
                                smooth[equal[0]])/(len(equal)+5)*(count+1)
            except BaseException:
                # we have last element
                dx = (smooth[1]-smooth[0])/10
                smooth[-1] += dx
    else:
        raise ValueError("Reached `max_iter`={}".format(max_iter))

    return smooth



"""Methods for estimating the point of contact (POC)"""
import lmfit
import numpy as np
from scipy.ndimage import uniform_filter1d


#: List of all methods available for contact point estimation
POC_METHODS = []


def compute_preproc_clip_approach(force):
    """Clip the approach part (discard the retract part)

    This POC preprocessing method may be applied before
    applying the POC estimation method.
    """
    # get data
    fg0 = np.array(force, copy=True)
    # Only use the (initial) approach part of the curve.
    idmax = np.argmax(fg0)
    fg = fg0[:idmax]
    return fg


def compute_poc(force, method="deviation_from_baseline", ret_details=False):
    """Compute the contact point from force data

    Parameters
    ----------
    force: 1d ndarray
        Force data
    method: str
        Name of the method for computing the POC (see :const:`POC_METHODS`)
    ret_details: bool
        Whether or not to return a dictionary with details alongside the
        POC estimate.

    Notes
    -----
    If the POC method returns np.nan, then the center of the
    force data is returned (to allow fitting algorithms to proceed).
    """
    # compute POC according to method chosen
    for mfunc in POC_METHODS:
        if mfunc.identifier == method:
            if "clip_approach" in mfunc.preprocessing:
                force = compute_preproc_clip_approach(force)
            data = mfunc(force, ret_details=ret_details)
            if ret_details:
                cp, details = data
                details["method"] = method
            else:
                cp, details = data, None
            break
    else:
        raise ValueError(f"Undefined POC method '{method}'!")
    if np.isnan(cp):
        cp = force.size // 2
    if ret_details:
        return cp, details
    else:
        return cp


def poc(identifier, name, preprocessing):
    """Decorator for point of contact (POC) methods

    The name and identifier are stored as a property of the wrapped
    function.

    Parameters
    ----------
    identifier: str
        identifier of the POC method (e.g. "baseline_deviation")
    name: str
        human-readble name of the POC method
        (e.g. "Deviation from baseline")
    preprocessing: list of str
        list of preprocessing methods that should be applied;
        may contain ["clip_approach"].
    """
    def attribute_setter(func):
        """Decorator that sets the necessary attributes

        The outer decorator is used to obtain the attributes.
        This inner decorator returns the actual function that
        wraps the preprocessor.
        """
        POC_METHODS.append(func)
        func.identifier = identifier
        func.name = name
        func.preprocessing = preprocessing
        return func

    return attribute_setter


@poc(identifier="deviation_from_baseline",
     name="Deviation from baseline",
     preprocessing=["clip_approach"])
def poc_deviation_from_baseline(force, ret_details=False):
    """Deviation from baseline

    1. Obtain the baseline (initial 10% of the gradient curve)
    2. Compute average and maximum deviation of the baseline
    3. The CP is the index of the curve where it exceeds
       twice of the maximum deviation
    """
    cp = np.nan
    details = {}
    # Crop the slow approach trace (10% of the curve)
    baseline = force[:int(force.size * .1)]
    if baseline.size:
        bl_avg = np.average(baseline)
        bl_rng = np.max(np.abs(baseline - bl_avg)) * 2
        bl_dev = (force - bl_avg) > bl_rng
        # argmax gets the first largest value
        maxid = np.argmax(bl_dev)
        if bl_dev[maxid]:
            cp = maxid
        if ret_details:
            x = [0, force.size-1]
            details["plot force"] = [np.arange(force.size), force]
            details["plot baseline mean"] = [x, [bl_avg, bl_avg]]
            details["plot baseline threshold"] = [x, [bl_avg + bl_rng,
                                                      bl_avg + bl_rng]]
            details["plot poc"] = [[cp, cp],
                                   [force.min(), force.max()]]

    if ret_details:
        return cp, details
    else:
        return cp


@poc(identifier="fit_constant_line",
     name="Piecewise fit with constant and line",
     preprocessing=["clip_approach"])
def poc_fit_constant_line(force, ret_details=False):
    r"""Piecewise fit with constant and line

    Fit a piecewise function (constant+linear) to the baseline
    and indentation part:

    .. math::

       F = \text{max}(d, m\delta + d)

    The point of contact is the intersection of a horizontal line
    at :math:`d` (baseline) and a linear function with slope :math:`m`
    for the indentation part.

    The point of contact is defined as :math:`\delta=0` (It's another
    fitting parameter).
    """
    def model(params, x):
        d = params["d"]
        x0 = params["x0"]
        m = params["m"]
        one = d
        two = m * (x - x0) + d
        return np.maximum(one, two)

    def residual(params, x, data):
        return data - model(params, x)

    y = np.array(force, copy=True)
    cp = np.nan
    details = {}
    if y.size > 4:  # 3 fit parameters
        ymin, ymax = np.min(y), np.max(y)
        y = (y - ymin) / (ymax - ymin)
        x = np.arange(y.size)
        x0 = poc_frechet_direct_path(force)
        if np.isnan(x0):
            x0 = y.size // 2
        params = lmfit.Parameters()
        params.add('d', value=np.mean(y[:10]))
        params.add('x0', value=x0)
        params.add('m', value=(1 - y[x0])/(x.size - x0))

        out = lmfit.minimize(residual, params, args=(x, y), method="nelder")
        if out.success:
            cp = int(out.params["x0"])
            if ret_details:
                details["plot force"] = [x, y]
                details["plot fit"] = [np.arange(force.size),
                                       model(out.params, x)]
                details["plot poc"] = [[cp, cp],
                                       [y.min(), y.max()]]

    if ret_details:
        return cp, details
    else:
        return cp


@poc(identifier="fit_constant_polynomial",
     name="Piecewise fit with constant and polynomial",
     preprocessing=["clip_approach"])
def poc_fit_constant_polynomial(force, ret_details=False):
    r"""Piecewise fit with constant and polynomial

    Fit a piecewise function (constant + polynomial) to the baseline
    and indentation part.

    .. math::

       F = \frac{\delta^3}{a\delta^2 + b\delta + c} + d

    This function is defined for all :math:`\delta>0`. For all
    :math:`\delta<0` the model evaluates to :math:`d` (baseline).

    I'm not sure where this has been described initially, but it is
    used e.g. in :cite:`Rusaczonek19`.

    For small indentations, this function exhibits a cubic behavior:

    .. math::

       y \approx \delta^3/c + d

    And for large indentations, this function is linear:

    .. math::

       y \approx \delta/a

    The point of contact is defined as :math:`\delta=0` (It's another
    fitting parameter).
    """
    def model(params, x):
        d = params["d"].value
        x0 = params["x0"].value
        a = params["a"].value
        b = params["b"].value
        c = params["c"].value
        x1 = x - x0
        curve = x1**3 / (a*x1**2 + b*x1 + c) + d
        curve[x1 <= 0] = d
        return curve

    def residual(params, x, data):
        curve = model(params, x)
        return data - curve

    y = np.array(force, copy=True)
    cp = np.nan
    details = {}
    if y.size > 6:  # 5 fit parameters
        ymin, ymax = np.min(y), np.max(y)
        y = (y - ymin) / (ymax - ymin)
        x = np.arange(y.size)
        x0 = poc_frechet_direct_path(force)
        if np.isnan(x0):
            x0 = y.size // 2
        params = lmfit.Parameters()
        params.add('d', value=np.mean(y[:10]))
        params.add('x0', value=x0)
        # The polynomial fitting parameters are supposed to be
        # greater than zero (source?). We set the minimum to 1e-3 so
        # the fitting algorithm becomes more stable. Also, the initial
        # values for b and c are more or less arbitrary (this is a heuristic
        # approach).
        # for larger x, a is something like an inverse slope. Since we
        # normalized the y-values to 1, we just take the x-difference.
        params.add('a', value=(y.size-x0), min=1e-3, max=100*(y.size-x0))
        params.add('b', value=y.size, min=1e-3)
        params.add('c', value=.5, min=1e-3)

        out = lmfit.minimize(residual, params, args=(x, y), method="nelder")

        if out.success:
            cp = int(out.params["x0"])
            if ret_details:
                details["plot force"] = [x, y]
                details["plot fit"] = [np.arange(force.size),
                                       model(out.params, x)]
                details["plot poc"] = [[cp, cp],
                                       [y.min(), y.max()]]

    if ret_details:
        return cp, details
    else:
        return cp


@poc(identifier="fit_line_polynomial",
     name="Piecewise fit with line and polynomial",
     preprocessing=["clip_approach"])
def poc_fit_line_polynomial(force, ret_details=False):
    r"""Piecewise fit with line and polynomial

    Fit a piecewise function (line + polynomial) to the baseline
    and indentation part.

    The linear basline (:math:`\delta<0`) is modeled with:

    .. math::

       F = m \delta + d

    The indentation part (:math:`\delta>0`) is modeled with:

    .. math::

       F = \frac{\delta^3}{a\delta^2 + b\delta + c} + m \delta  + d

    For small indentations, this function exhibits a linear and
    only slightly cubic behavior:

    .. math::

       y \approx \delta^3/c + m \delta + d

    And for large indentations, this function is linear:

    .. math::

       y \approx \left( \frac{1}{a} + m \right) \delta

    The point of contact is defined as :math:`\delta=0` (It's another
    fitting parameter).


    See Also
    --------
    poc_fit_constant_polynomial: polynomial-only version
    """
    def model(params, x):
        d = params["d"].value
        x0 = params["x0"].value
        m = params["m"].value
        a = params["a"].value
        b = params["b"].value
        c = params["c"].value
        x1 = x - x0
        curve = m * x1 + d
        curve[x1 > 0] += x1[x1 > 0]**3 / (a*x1[x1 > 0]**2 + b*x1[x1 > 0] + c)
        return curve

    def residual(params, x, data):
        curve = model(params, x)
        return data - curve

    y = np.array(force, copy=True)
    cp = np.nan
    details = {}
    if y.size > 7:  # 6 fit parameters
        ymin, ymax = np.min(y), np.max(y)
        y = (y - ymin) / (ymax - ymin)
        x = np.arange(y.size)
        x0 = poc_frechet_direct_path(force)
        if np.isnan(x0):
            x0 = y.size // 2
        params = lmfit.Parameters()
        params.add('d', value=np.mean(y[:10]))
        params.add('x0', value=x0)
        # slope
        params.add('m', value=y[x0]/x0)
        # The polynomial fitting parameters are supposed to be
        # greater than zero (source?). We set the minimum to 1e-3 so
        # the fitting algorithm becomes more stable. Also, the initial
        # values for b and c are more or less arbitrary (this is a heuristic
        # approach).
        # for larger x, a is something like an inverse slope. Since we
        # normalized the y-values to 1, we just take the x-difference.
        params.add('a', value=(y.size-x0), min=1e-3, max=100*(y.size-x0))
        params.add('b', value=y.size, min=1e-3)
        params.add('c', value=.5, min=1e-3)

        out = lmfit.minimize(residual, params, args=(x, y), method="nelder")

        if out.success:
            cp = int(out.params["x0"])
            if ret_details:
                details["plot force"] = [x, y]
                details["plot fit"] = [np.arange(force.size),
                                       model(out.params, x)]
                details["plot poc"] = [[cp, cp],
                                       [y.min(), y.max()]]
                print(cp, y.min(), y.max())

    if ret_details:
        return cp, details
    else:
        return cp


@poc(identifier="frechet_direct_path",
     name="Fréchet distance to direct path",
     preprocessing=["clip_approach"])
def poc_frechet_direct_path(force, ret_details=False):
    """Fréchet distance to direct path

    The indentation part is transformed to normalized coordinates
    (force and corresponding x in range [0, 1]). The point with the
    largest distance to the line from (0, 0) to (1, 1) is the contact
    point.

    This method is robust with regard to tilted baselines and is a
    good initial guess for fitting-based POC estimation approaches.

    Note that the length of the baseline influences the returned
    contact point. For shorter baselines, the contact point will
    be closer to the point of maximum indentation.
    """
    x = np.linspace(0, 1, len(force), endpoint=True)
    y = (force - force.min()) / (force.max() - force.min())

    # rotate the curve towards x
    # (computation of Frechet distance with curve is now just distance
    # from x-axis, i.e. minimum)
    alpha = - np.pi / 4
    yr = x * np.sin(alpha) + y * np.cos(alpha)
    cp = np.argmin(yr)

    if ret_details:
        details = {"plot normalized rotated force": [np.arange(len(force)),
                                                     yr],
                   "plot poc": [[cp, cp],
                                [yr.min(), yr.max()]]}
        return cp, details
    else:
        return cp


@poc(identifier="gradient_zero_crossing",
     name="Gradient zero-crossing of indentation part",
     preprocessing=["clip_approach"])
def poc_gradient_zero_crossing(force, ret_details=False):
    """Gradient zero-crossing of indentation part

    1. Apply a moving average filter to the curve
    2. Compute the gradient
    3. Cut off gradient at maximum with a 10 point reserve
    4. Apply a moving average filter to the gradient
    5. The POC is the index of the averaged gradient curve where
       the values are below 1% of the gradient maximum, measured
       from the indentation maximum (not from baseline).
    """
    cp = np.nan
    details = {}
    # Perform a median filter to smooth the array
    filtsize = max(5, int(force.size*.01))
    y = uniform_filter1d(force, size=filtsize)
    if y.size > 1:
        # Cutoff at maximum plus some reserve
        cutoff = y.size - np.argmax(y) + 10
        grad = np.gradient(y)[:-cutoff]
        if grad.size > 50:
            # Use the point where the gradient becomes small enough.
            gradn = uniform_filter1d(grad, size=filtsize)
            thresh = 0.01 * np.max(gradn)
            gradpos = gradn <= thresh
            if np.sum(gradpos):
                # The gradient contains positive values.
                # Flip `gradpos`, because we want the first value from the
                # end of the array.
                # Weight with two times "filtsize//2", because we actually
                # want the rolling median filter from the edge and not at the
                # center of the array (and two times, because we did two
                # filter operations).
                cp = y.size - np.where(gradpos[::-1])[0][0] - cutoff + filtsize

                if ret_details:
                    x = np.arange(gradn.size)
                    details["plot force gradient"] = [x, gradn]
                    details["plot threshold"] = [[x[0], x[-1]],
                                                 [thresh, thresh]]
                    details["plot poc"] = [[cp, cp],
                                           [gradn.min(), gradn.max()]]

    if ret_details:
        return cp, details
    else:
        return cp



import importlib
import pathlib
import sys
import warnings

from .core import ModelImportError, NaniteFitModel

#: currently available models
models_available = {}


def load_model_from_file(path, register=False):
    """Import a fit model file and return the module

    This is intended for loading custom models or for model
    development.

    Parameters
    ----------
    path: str or Path
        pathname to a Python script conaining a fit model
    register: bool
        whether to register the model after import

    Returns
    -------
    model: NaniteFitModel
        nanite fit model object

    Raises
    ------
    ModelImportError
        If the model cannot be imported
    """
    path = pathlib.Path(path)
    try:
        # insert the plugin directory to sys.path so we can import it
        sys.path.insert(-1, str(path.parent))
        sys.dont_write_bytecode = True
        module = importlib.import_module(path.stem)
    except ModuleNotFoundError:
        raise ModelImportError(f"Could not import '{path}'!")
    finally:
        # undo our path insertion
        sys.path.pop(0)
        sys.dont_write_bytecode = False

        mod = NaniteFitModel(module)

        if register:
            register_model(module)

        return mod


def register_model(module, *args):
    """Register a fitting model

    Parameters
    ----------
    module: Python module or NaniteFitModel
        the model to register

    Returns
    -------
    model: NaniteFitModel
        the corresponding NaniteFitModel instance
    """
    if args:
        warnings.warn("Please only pass the module to `register_model`!",
                      DeprecationWarning)
    global models_available  # this is not necessary, but clarifies things
    # add model
    if isinstance(module, NaniteFitModel):
        # we already have a fit model
        md = module
    else:
        md = NaniteFitModel(module)
    # the actual registration
    models_available[module.model_key] = md
    return md


def deregister_model(model):
    """Deregister a NaniteFitModel"""
    global models_available  # this is not necessary, but clarifies things
    models_available.pop(model.model_key)



import pathlib
import shutil
import tempfile

import nanite.read


data_path = pathlib.Path(__file__).parent / "data"


def test_recursive_callback():
    # set up a directory with recursive data
    td = pathlib.Path(tempfile.mkdtemp(prefix="nanite_recursive_load_"))
    td2 = td / "data1"
    td2.mkdir(parents=True)
    td3 = td / "data3" / "data4"
    td3.mkdir(parents=True)

    shutil.copy2(data_path / "fmt-jpk-fd_spot3-0192.jpk-force",
                 td2 / "spot1.jpk-force")
    shutil.copy2(data_path / "fmt-jpk-fd_spot3-0192.jpk-force",
                 td2 / "spot2.jpk-force")
    shutil.copy2(data_path / "fmt-jpk-fd_spot3-0192.jpk-force",
                 td3 / "spot3.jpk-force")

    # trace the callback calls
    calls = []

    def mycallback(value):
        calls.append(value)

    files = nanite.read.load_data(td, callback=mycallback)
    assert len(files) == 3

    assert calls == [1/3, 2/3, 1]



"""Test of data set functionalities"""
import pathlib

import lmfit
import numpy as np

from nanite import IndentationGroup
from nanite.model import model_sneddon_spherical_approximation as mod_ssa


data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_app_ret():
    ds = IndentationGroup(jpkfile)
    ar = ds[0]
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset"])
    idp = ar.estimate_contact_point_index()

    aprid = ar["segment"] == 0
    x = ar["tip position"][aprid]
    y = ar["force"][aprid]
    contact_point = x[idp]

    # crop x and y around contact_point
    distcp = x.shape[0]-idp
    x = x[-3*distcp:]
    y = y[-3*distcp:]

    params = lmfit.Parameters()
    params.add("contact_point", value=contact_point)
    params.add("baseline", value=0, vary=False)
    params.add("E", value=300e3, min=0)
    params.add("nu", value=.5, vary=False)
    params.add("R", value=40e-9, vary=False)

    fit_w = lmfit.minimize(mod_ssa.residual, params, args=(x, y, True))
    fit_n = lmfit.minimize(mod_ssa.residual, params, args=(x, y, False))

    if __name__ == "__main__" and False:
        import matplotlib.pylab as plt
        _fig, axes = plt.subplots(2, 1)
        xf = np.linspace(x[0], x[-1], 100)
        axes[0].plot(xf, mod_ssa.model(fit_w.params, xf),
                     label="fit with weights")
        axes[0].plot(xf, mod_ssa.model(fit_n.params, xf),
                     label="fit no weights")
        axes[0].plot(x, y, label="data")
        axes[0].legend()
        axes[0].grid()

        axes[1].plot(x, mod_ssa.residual(fit_w.params, x, y, weight_cp=True),
                     label="residuals with weight")
        axes[1].plot(x, mod_ssa.residual(fit_n.params, x, y, weight_cp=False),
                     label="residuals no weight")
        axes[1].legend()
        axes[1].grid()

        plt.show()


def test_fit_apret():
    ds = IndentationGroup(jpkfile)
    ar = ds[0]
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset"])
    ar.fit_model(model_key="sneddon_spher_approx",
                 params_initial=None,
                 x_axis="tip position",
                 y_axis="force",
                 weight_cp=False,
                 segment="retract")

    if __name__ == "__main__":
        import matplotlib.pylab as plt
        _fig, axes = plt.subplots(2, 1)
        axes[0].plot(ar["tip position"], ar["force"], label="data")
        axes[0].plot(ar["tip position"], ar["fit"], label="fit")
        axes[0].legend()
        axes[0].grid()
        plt.show()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test hashing of fit results"""
import pathlib
import time

import nanite
from nanite import IndentationGroup


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_hash_time():
    ds1 = IndentationGroup(jpkfile)
    apret = ds1[0]
    apret.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8321e-5)

    # Fit with absolute full range
    kwargs = dict(model_key="hertz_para",
                  params_initial=inparams,
                  range_x=(0, 0),
                  range_type="absolute",
                  x_axis="tip position",
                  y_axis="force",
                  segment="approach",
                  weight_cp=False)
    t0 = time.perf_counter()
    apret.fit_model(**kwargs)
    t1 = time.perf_counter()
    apret.fit_model()
    t2 = time.perf_counter()
    kwargs["weight_cp"] = 1e-5
    apret.fit_model(**kwargs)
    t3 = time.perf_counter()
    apret.fit_model()
    t4 = time.perf_counter()

    assert t1-t0 >= 100 * \
        (t2-t1), "Consecutive fits with same parameters should be instant"
    assert t3-t2 >= 100 * \
        (t2-t1), "Changing parameters again should cause a new fit"
    assert t3-t2 >= 100*(t4-t3), "And computing the same should be faster"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from collections import OrderedDict
import copy
import warnings

import afmformats
import numpy as np

from .fit import IndentationFitter, FitProperties, guess_initial_parameters, \
    FP_DEFAULT
from . import model
from . import poc
from . import preproc
from .rate import get_rater


class Indentation(afmformats.AFMForceDistance):
    def __init__(self, data, metadata, diskcache=None):
        """Additional functionalities for afmformats.AFMForceDistance"""
        super(Indentation, self).__init__(data=data,
                                          metadata=metadata,
                                          diskcache=diskcache)
        #: Default preprocessing steps,
        #: see :func:`Indentation.apply_preprocessing`.
        self.preprocessing = []
        #: Preprocessing options
        self.preprocessing_options = {}
        # preprocessing details (for user convenience)
        self._preprocessing_details = {}
        # protected fit properties
        self._fit_properties = FitProperties()

        # Curve rating (see `self.rate_quality`)
        self._rating = None

    @property
    def data(self):
        warnings.warn("Please use __getitem__ instead!", DeprecationWarning)
        return self

    @property
    def fit_properties(self):
        """Fitting results, see :func:`Indentation.fit_model`)"""
        return self._fit_properties

    @fit_properties.setter
    def fit_properties(self, fp):
        self._fit_properties.update(fp)

    def apply_preprocessing(self, preprocessing=None, options=None,
                            ret_details=False):
        """Perform curve preprocessing steps

        Parameters
        ----------
        preprocessing: list
            A list of preprocessing method identifiers that are
            stored in the `nanite.preproc.PREPROCESSORS` list.
            If set to `None`, `self.preprocessing` will be
            used.
        options: dict of dict
            Dictionary of keyword arguments for each preprocessing
            step (if applicable)
        ret_details:
            Return preprocessing details dictionary
        """
        if preprocessing is None:
            preprocessing = self.preprocessing

        if options is None:
            options = self.preprocessing_options

        if "preprocessing" in self.fit_properties:
            preproc_past = [self.fit_properties["preprocessing"],
                            self.fit_properties["preprocessing_options"]]
        else:
            preproc_past = []

        if ((preproc_past != [preprocessing, options])
                or (not self._preprocessing_details and ret_details)):
            # Remember initial fit parameters for user convenience
            fp = self.fit_properties
            # Reset fit properties
            fp.reset()
            # Set preprocessing options
            fp["preprocessing"] = preprocessing
            fp["preprocessing_options"] = options
            # Reset rating
            self._rating = None
            # Apply preprocessing
            # (This will call `AFMData.reset_data` on self)
            details = preproc.apply(apret=self,
                                    identifiers=preprocessing,
                                    options=options,
                                    ret_details=ret_details)
            self._preprocessing_details = details
            # Check availability of axes
            for ax in ["x_axis", "y_axis"]:
                # make sure the fitting axes are defined
                if ax in fp and not fp[ax] in self:
                    fp.pop(ax)

        # remember preprocessing
        self.preprocessing = preprocessing
        self.preprocessing_options = copy.deepcopy(options)

        return self._preprocessing_details

    def compute_emodulus_mindelta(self, callback=None):
        """Elastic modulus in dependency of maximum indentation

        The fitting interval is varied such that the maximum
        indentation depth ranges from the lowest tip position
        to the estimated contact point. For each interval, the
        current model is fitted and the elastic modulus is
        extracted.

        Parameters
        ----------
        callback: callable
            A method that is called with the `emoduli` and
            `indentations` as the computation proceeds every
            five steps.

        Returns
        -------
        emoduli, indentations: 1d ndarrays
            The fitted elastic moduli at the corresponding
            maximal indentation depths.

        Notes
        -----
        The information about emodulus and mindelta is also stored in
        `self.fit_properties` with the keys "optimal_fit_E_array" and
        "optimal_fit_delta_array", if `self.fit_model` is called with
        the argument `search_optimal_fit` set to `True`.
        """

        if "optimal_fit_E_array" in self.fit_properties:
            emoduli = self.fit_properties["optimal_fit_E_array"]
            indentations = self.fit_properties["optimal_fit_delta_array"]
        else:
            fitter = IndentationFitter(self)
            emoduli, indentations = fitter.compute_emodulus_vs_mindelta(
                callback=callback
            )
            self.fit_properties["optimal_fit_E_array"] = emoduli
            self.fit_properties["optimal_fit_delta_array"] = indentations
        return emoduli, indentations

    def estimate_optimal_mindelta(self):
        """Estimate the optimal indentation depth

        This is a convenience function that wraps around
        `compute_emodulus_mindelta` and
        `IndentationFitter.compute_opt_mindelta`.
        """
        emoduli, indentations = self.compute_emodulus_mindelta()
        dopt = IndentationFitter.compute_opt_mindelta(
            emoduli=emoduli,
            indentations=indentations
        )
        return dopt

    def estimate_contact_point_index(self, method="deviation_from_baseline"):
        """Estimate the contact point index

        See the `poc` submodule for more information.
        """
        idp = poc.compute_poc(force=np.array(self["force"], copy=True),
                              method=method)
        return idp

    def fit_model(self, **kwargs):
        """Fit the approach-retract data to a model function

        Parameters
        ----------
        model_key: str
            A key referring to a model in
            `nanite.model.models_available`
        params_initial: instance of lmfit.Parameters or dict
            Parameters for fitting. If not given,
            default parameters are used.
        range_x: tuple of 2
            The range for fitting, see `range_type` below.
        range_type: str
            One of:

            - absolute:
                Set the absolute fitting range in values
                given by the `x_axis`.
            - relative cp:
                In some cases it is desired to be able to
                fit a model only up until a certain indentation
                depth (tip position) measured from the contact
                point. Since the contact point is a fit parameter
                as well, this requires a two-pass fitting.
        preprocessing: list of str
            Preprocessing
        preprocessing_options: list of str
            Preprocessing
        segment: str
            Segment index (e.g. 0 for approach)
        weight_cp: float
            Weight the contact point region which shows artifacts
            that are difficult to model with e.g. Hertz.
        gcf_k: float
            Geometrical correction factor :math:`k` for non-single-contact
            data. The measured indentation is multiplied by this factor to
            correct for experimental geometries during fitting,
            e.g. ``gcf_k=0.5`` for parallel-place compression.
        optimal_fit_edelta: bool
            Search for the optimal fit by varying the maximal
            indentation depth and determining a plateau in the
            resulting Young's modulus (fitting parameter "E").
        """
        if "preprocessing" in kwargs:
            options = kwargs.get("preprocessing_options",
                                 self.preprocessing_options)
            self.apply_preprocessing(preprocessing=kwargs["preprocessing"],
                                     options=options)
        # self.fit_properties is an instance of FitProperties that
        # stores previous fit kwargs. If the given kwargs are
        # different than in the previous fit, the following two
        # lines will reset the "hash" in the fit properties, triggering
        # a new fit.
        # (sorted, such that `model_key` is set before `params_initial`)
        for arg in sorted(kwargs.keys()):
            self.fit_properties[arg] = kwargs[arg]

        # set a default model (needed for self.get_initial_fit_parameters)
        if "model_key" not in self.fit_properties:
            self.fit_properties["model_key"] = FP_DEFAULT["model_key"]

        # set default initial parameters
        if ("params_initial" not in self.fit_properties
                or self.fit_properties["params_initial"] is None):
            # We need the initial parameters (to modify them).
            # Guesses common parameters like the contact point that
            # would have otherwise been done in `IndentationFitter`:
            fp_guess = self.get_initial_fit_parameters(
                common_ancillaries=True,
                model_ancillaries=True)
            self.fit_properties["params_initial"] = fp_guess

        if "hash" in self.fit_properties:
            # There is nothing to do, because the initial fit
            # properties are the same.
            pass
        else:
            fitter = IndentationFitter(self)
            # Perform fitting
            # Note: if `fitter.fp["success"]` is `False`, then
            # the `fit_residuals` and `fit_curve` are `nan`.
            fitter.fit()
            self["fit"] = fitter.fit_curve
            self["fit residuals"] = fitter.fit_residuals
            self["fit range"] = fitter.fit_range
            self.fit_properties = fitter.fp

    def get_ancillary_parameters(self, model_key=None):
        """Compute ancillary parameters for the current model"""
        if model_key is None:
            if "model_key" in self.fit_properties:
                model_key = self.fit_properties["model_key"]
            else:
                model_key = FP_DEFAULT["model_key"]
        return model.compute_anc_parms(idnt=self,
                                       model_key=model_key)

    def get_initial_fit_parameters(self, model_key=None,
                                   common_ancillaries=True,
                                   model_ancillaries=True):
        """Return the initial fit parameters

        If there are not initial fit parameters set in
        `self.fit_properties`, then they are computed.

        Parameters
        ----------
        model_key: str
            Optionally set a model key. This will override the
            "model_key" key in `self.fit_properties`.
        common_ancillaries: bool
            Guess global ancillaries such as the contact point.
        model_ancillaries: bool
            Guess model-related ancillaries

        Notes
        -----
        `global_ancillaries` and `model_ancillaries` only have an
        effect if self.fit_properties["params_initial"] is set.
        """
        if model_key is not None:
            self.fit_properties["model_key"] = model_key
        if self.fit_properties.get("params_initial", False):
            parms = self.fit_properties["params_initial"]
        elif "model_key" in self.fit_properties:
            parms = guess_initial_parameters(
                self,
                model_key=self.fit_properties["model_key"],
                model_ancillaries=model_ancillaries,
                common_ancillaries=common_ancillaries)
        else:
            # for user convenience (with default model)
            parms = guess_initial_parameters(
                self,
                model_key=FP_DEFAULT["model_key"],
                model_ancillaries=model_ancillaries,
                common_ancillaries=common_ancillaries)
        return parms

    def get_rating_parameters(self):
        """Return current rating parameters"""
        rdict = OrderedDict()
        if self._rating is None:
            rt = [np.nan] * 6
        else:
            rt = self._rating
        rdict["Hash"] = rt[0]
        rdict["Regressor"] = rt[1]
        rdict["Training set"] = rt[2]
        rdict["Feature names"] = rt[3]
        rdict["Linear discriminant analysis"] = rt[4]
        rdict["Rating"] = rt[5]
        return rdict

    def rate_quality(self, regressor="Extra Trees", training_set="zef18",
                     names=None, lda=None):
        """Compute the quality of the obtained curve

        Uses heuristic approaches to rate a curve.

        Parameters
        ----------
        regressor: str
            The regressor name used for rating.
        training_set: str
            A label for a training set shipped with nanite or a
            path to a training set.
        names: list of str
            Only use these features for rating
        lda: bool
            Perform linear discriminant analysis

        Returns
        -------
        rating: float
            A value between 0 and 10 where 0 is the lowest rating.
            If no fit has been performed, a rating of -1 is returned.

        Notes
        -----
        The rating is cached based on the fitting hash
        (see `IndentationFitter._hash`).
        """
        if self.fit_properties and "hash" in self.fit_properties:
            curhash = self.fit_properties["hash"]
        else:
            curhash = "none"
        if regressor.lower() == "none":
            rt = -1
        elif (self._rating is None or
              self._rating[0] != curhash or
              self._rating[1] != regressor or
              self._rating[2] != training_set or
              self._rating[3] != names or
              self._rating[4] != lda):
            # Perform rating
            rater = get_rater(regressor=regressor,
                              training_set=training_set,
                              names=names,
                              lda=lda)
            rt = rater.rate(datasets=self)[0]
            self._rating = (curhash, regressor, training_set, names, lda, rt)
        else:
            # Use cached rating
            rt = self._rating[-1]
        return rt



import pathlib
import tempfile

import numpy as np

from nanite.cli import profile, rating
from nanite.rate import IndentationRater


data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
jpkfile2 = data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map"


def setup_training_set(n=300):
    tdir = tempfile.mkdtemp(prefix="test_nanite_rate_ts_")
    tdir = pathlib.Path(tdir)
    np.random.set_state(np.random.RandomState(47).get_state())
    for bb in IndentationRater.get_feature_names(which_type="binary"):
        bvals = np.random.choice([0, 1], size=n, p=[.05, .95])
        np.savetxt(tdir / "train_{}.txt".format(bb), bvals)
    for cc in IndentationRater.get_feature_names(which_type="continuous"):
        cvals = np.random.random_sample(size=n)
        np.savetxt(tdir / "train_{}.txt".format(cc), cvals)
    thisrating = np.random.choice(range(11), size=n)
    np.savetxt(tdir / "train_response.txt", thisrating)
    return tdir


def test_fit_data():
    # use temporary file
    _, name = tempfile.mkstemp(suffix=".cfg", prefix="test_nanite_cli_rate_")
    name = pathlib.Path(name)
    profile.Profile(path=name)

    # this will fit with the profile default parameters
    idnt = rating.fit_data(path=jpkfile, profile_path=name)
    assert idnt.path == jpkfile
    assert idnt.fit_properties["success"]


def test_fit_data_with_user_training_set():
    tdir = setup_training_set()
    _, name = tempfile.mkstemp(suffix=".cfg", prefix="test_nanite_cli_rate_")
    name = pathlib.Path(name)
    pf = profile.Profile(path=name)
    pf["rating training set"] = tdir
    pf["fit param R value"] = 37.28e-6 / 2
    pout = tempfile.mkdtemp(prefix="test_nanite_cli_rate_ts")
    pout = pathlib.Path(pout)
    rating.fit_perform(path=jpkfile2, path_results=pout, profile_path=name)
    stats = np.loadtxt(pout / "statistics.tsv", skiprows=1, usecols=(1, 2, 3))
    assert np.all(stats[:, 0] == [109, 129, 416])
    assert np.all((3.5 < stats[:, 2]) * (stats[:, 2] < 5))


def test_fit_data_with_zef18():
    _, name = tempfile.mkstemp(suffix=".cfg", prefix="test_nanite_cli_rate_")
    name = pathlib.Path(name)
    pf = profile.Profile(path=name)
    pf["rating training set"] = "zef18"
    pf["weight_cp"] = 2e-6
    pf["fit param R value"] = 137.28e-6 / 2
    pout = tempfile.mkdtemp(prefix="test_nanite_cli_rate_ts")
    pout = pathlib.Path(pout)
    rating.fit_perform(path=jpkfile2, path_results=pout, profile_path=name)
    stats = np.loadtxt(pout / "statistics.tsv", skiprows=1, usecols=(1, 2, 3))
    assert np.all(stats[:, 0] == [109, 129, 416])
    assert stats[0, 2] == 9.5
    assert stats[1, 2] == 2.4
    assert stats[2, 2] == 4.9


if __name__ == "__main__":
    # Run all tests
    test_fit_data_with_zef18()
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



# flake8: noqa: F401
from .group import IndentationGroup, load_group
from .indent import Indentation
from . import model
from .qmap import QMap
from .rate import IndentationRater

from ._version import version as __version__



import pathlib

import afmformats
from afmformats.errors import MissingMetaDataError

from .read import get_load_data_modality_kwargs, load_data


def load_group(path, callback=None, meta_override=None):
    """Load indentation data from disk

    Parameters
    ----------
    path: path-like
        Path to experimental data
    callback: callable
        function for tracking progress; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        should be used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)

    Returns
    -------
    group: nanite.IndetationGroup
        Indentation group with force-distance data
    """
    path = pathlib.Path(path)
    data = load_data(path,
                     callback=callback,
                     meta_override=meta_override,
                     )
    grp = IndentationGroup()
    grp += data
    grp.path = path
    return grp


class IndentationGroup(afmformats.AFMGroup):
    def __init__(self, path=None, meta_override=None, callback=None):
        """Group of Indentation

        Parameters
        ----------
        path: str or pathlib.Path or None
            The path to the data file. The data format is determined
            and the file is loaded using :ref:`afmformats:index`.
        meta_override: dict
            if specified, contains key-value pairs of metadata that
            should be used when loading the files
            (see :data:`afmformats.meta.META_FIELDS`)
        callback: callable or None
            A method that accepts a float between 0 and 1
            to externally track the process of loading the data.
        """
        super(IndentationGroup, self).__init__(
            path=path,
            meta_override=meta_override,
            callback=callback,
            **get_load_data_modality_kwargs()
        )

    def append(self, afmdata):
        """Append a new instance of AFMData

        This subclassed method makes sure that "spring constant" is set
        if "tip position" has to be computed in the future.

        Parameters
        ----------
        afmdata: afmformats.afm_data.AFMData
            AFM data
        """
        if ("spring constant" not in afmdata.metadata
                and "tip position" not in afmdata):
            raise MissingMetaDataError(
                ["spring constant"],
                "Please specify the spring constant!")
        # Call the original function
        super(IndentationGroup, self).append(afmdata)



"""Test of model basics"""
import warnings

import nanite
import nanite.model

from common import MockModelModule


def test_bad_parameter_order():
    swapped_keys = ["R", "E", "nu", "contact_point", "baseline"]
    mod = MockModelModule("test_bad_order", parameter_keys=swapped_keys)
    try:
        mod.__enter__()
    except nanite.model.core.ModelImplementationError:
        pass
    else:
        assert False, "Bad parameter order should not be possible"
    finally:
        try:
            mod.__exit__(None, None, None)
        except KeyError:
            pass


def test_bad_parameter_number():
    short_keys = ["E", "R", "nu", "contact_point"]
    mod = MockModelModule("test_bad_number", parameter_keys=short_keys)
    try:
        mod.__enter__()
    except nanite.model.core.ModelImplementationError:
        pass
    else:
        assert False, "Bad parameter number should not be possible"
    finally:
        try:
            mod.__exit__(None, None, None)
        except KeyError:
            pass


def test_bad_parameter_units_number():
    short_units = ["Pa", "m", "", "m"]
    mod = MockModelModule("test_bad_units_number", parameter_units=short_units)
    try:
        mod.__enter__()
    except nanite.model.core.ModelImplementationError:
        pass
    else:
        assert False, "Bad parameter units number should not be possible"
    finally:
        try:
            mod.__exit__(None, None, None)
        except KeyError:
            pass


def test_bad_model_func_args():
    with warnings.catch_warnings():
        warnings.simplefilter("error",
                              nanite.model.core.ModelImplementationWarning)

        def bad_func(E, delta, R, nu, contact_point=0, baseline=0):
            return delta
        mod = MockModelModule("test_bad_func", model_func=bad_func)
        try:
            mod.__enter__()
        except nanite.model.core.ModelImplementationWarning:
            pass
        else:
            assert False, "Bad parameter order should not be possible"
        finally:
            try:
                mod.__exit__(None, None, None)
            except KeyError:
                pass


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
